from .Iscca import iscca
from .Iscca_cv import iscca_cv
from .Iscca_plot import iscca_plot
from .Preview_cca import preview_cca
from .Meta_scca import meta_scca
from .Scca import scca
from .Ispca import ispca
from .Ispca_cv import ispca_cv
from .Ispca_plot import ispca_plot
from .Preview_pca import preview_pca
from .Meta_spca import meta_spca
from .Spca import spca
from .Ispls import ispls
from .Ispls_cv import ispls_cv
from .Ispls_plot import ispls_plot
from .Preview_pls import preview_pls
from .Meta_spls import meta_spls
from .Spls import spls

__all__ = ['iscca', 'iscca_cv', 'iscca_plot', 'preview_cca','meta_scca','scca', 
           'ispca','ispca_cv','ispca_plot','preview_pca','meta_spca','spca',
           'ispls','ispls_cv','ispls_plot','preview_pls','meta_spls','spls']


import numpy as np
import pandas as pd
from scipy.integrate import quad
from sklearn.preprocessing import scale
from scipy.sparse.linalg import svds
import matplotlib.pyplot as plt

def iscca(x, y, L, mu1, mu2, mu3, mu4, eps=1e-04, pen1="homogeneity", 
          pen2="magnitude", scale_x=True, scale_y=True, maxstep=50, 
          submaxstep=10, trace=False, draw=False):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    if not isinstance(y, list):
        raise TypeError("y should be of list type.")
    x = [np.array(i) for i in x]
    y = [np.array(i) for i in y]  
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x]    
    ql = [i.shape[1] for i in y]  
    p = list(set(pl))
    q = list(set(ql))   
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(q) > 1:
        raise ValueError("The dimension of data y should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    if len(q) == 1:
        q = q[0] 
    ip = np.array(range(1, p+1))
    iq = np.array(range(1, q+1))
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    meany = [np.dot(np.ones((1, nl[l])), y[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
        y[l] = scale(y[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    def y_scale(l):
        one = np.ones((1, nl[l]))
        normy = np.sqrt(np.dot(one, y[l]**2) / (nl[l] - 1))
        if np.any(normy < np.finfo(float).eps):
            raise ValueError("Some of the columns of the response matrix have zero variance.")
        return normy
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_y:
        normy = [y_scale(i) for i in range(L)]
    else:
        normy = [[1]*q[0]]*L
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    if scale_y:
        y = [(y[i] - np.mean(y[i], axis=0)) / normy[i] for i in range(L)]    
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    def u_value_homo(u, v, p, q, L, mu1, mu2, pen2):
        def fun_s(j, l):
            s1 = np.dot(v[:, l].T, y[l].T).dot(x[l][:, j].reshape(1, -1).T) / nl[l]
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(u[j, np.arange(len(u[0]))!=l])
            if pen2 == "sign":
                s2 = mu2 * np.sum([x / np.sqrt(x**2 + 0.5) for x in u[j, np.arange(len(u[0]))!=l]]) / np.sqrt(u[j, l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)])
        s = result_s.reshape(p, L)
        norm_u_j = np.apply_along_axis(lambda x: np.sqrt(np.sum(x**2)), 1, u)
        ro_d = ro_d1st(norm_u_j, mu1, 6)
        def fun_c(j, l):
            s_norm = np.sqrt(np.sum(s[j, :]**2))
            if pen2 == "magnitude":
                c = (s_norm > ro_d[j]) * s[j, l] * (s_norm - ro_d[j]) / ((mu2 * (L - 1)) * s_norm)
            if pen2 == "sign":
                c = (s_norm > ro_d[j]) * s[j, l] * (s_norm - ro_d[j]) / ((mu2 * (L - 1) / (u[j, l]**2 + 0.5)) * s_norm)
            return c
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def u_value_hetero(u, v, p, q, L, mu1, mu2, pen2):
        def fun_mu(j, l):
            u_j = u[j, :]
            ro_j = ro(np.abs(u_j), mu1, 6)
            s_ro = np.sum(ro_j)
            mu_jl = ro_d1st(s_ro, 1, 1/2 * L * 6 * mu1**2) * ro_d1st(abs(u_j[l]), mu1, 6)
            return mu_jl
        result_mu = np.array([fun_mu(j, l) for j in range(p) for l in range(L)])
        mu = result_mu.reshape(p, L)
        def fun_s(j, l):
            s1 = np.dot(v[:, l].T, y[l].T).dot(x[l][:, j].reshape(1, -1).T) / nl[l]
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(u[j, np.arange(len(u[0]))!=l])
            if pen2 == "sign":
                s2 = mu2 * np.sum([x / np.sqrt(x**2 + 0.5) for x in u[j, np.arange(len(u[0]))!=l]]) / np.sqrt(u[j, l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)])
        s = result_s.reshape(p, L)
        def fun_c(j, l):
            if pen2 == "magnitude":
                c = np.sign(s[j, l]) * (abs(s[j, l]) > mu[j, l]) * (abs(s[j, l]) - mu[j, l]) / (mu2 * (L - 1))
            if pen2 == "sign":
                c = np.sign(s[j, l]) * (abs(s[j, l]) > mu[j, l]) * (abs(s[j, l]) - mu[j, l]) / (mu2 * (L - 1) / (u[j, l]**2 + 0.5))
            return c
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def v_value_homo(u, v, p, q, L, mu1, mu2, pen2):
        def fun_s(j, l):
            s1 = np.dot(u[:, l].T, x[l].T).dot(y[l][:, j].reshape(1, -1).T) / nl[l]
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(v[j, np.arange(L)!=l])
            if pen2 == "sign":
                s2 = mu2 * np.sum([x/np.sqrt(x**2 + 0.5) for x in v[j, np.arange(L)!=l]]) / np.sqrt(v[j, l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = [fun_s(j, l) for j in range(q) for l in range(L)]
        s = np.array(result_s).reshape(q, L)
        norm_v_j = np.sqrt(np.sum(v**2, axis=1))
        ro_d = ro_d1st(norm_v_j, mu1, 6)
        def fun_c(j, l):
            s_norm = np.sqrt(np.sum(s[j, :]**2))
            if pen2 == "magnitude":
                c = (s_norm > ro_d[j]) * s[j, l] * (s_norm - ro_d[j]) / ((mu2 * (L - 1)) * s_norm)
            if pen2 == "sign":
                c = (s_norm > ro_d[j]) * s[j, l] * (s_norm - ro_d[j]) / ((mu2 * (L - 1) / (v[j, l]**2 + 0.5)) * s_norm)
            return c
        c = np.array([fun_c(j, l) for j in range(q) for l in range(L)]).reshape(q, L)
        return c
    def v_value_hetero(u, v, p, q, L, mu1, mu2, pen2):
        def fun_mu(j, l):
            v_j = v[j, :]
            ro_j = ro(np.abs(v_j), mu1, 6)
            s_ro = np.sum(ro_j)
            mu_jl = ro_d1st(s_ro, 1, 1/2 * L * 6 * mu1**2) * ro_d1st(abs(v_j[l]), mu1, 6)
            return mu_jl
        result_mu = [fun_mu(j, l) for j in range(q) for l in range(L)]
        mu = np.array(result_mu).reshape(q, L)
        def fun_s(j, l):
            s1 = np.dot(u[:, l].T, x[l].T).dot(y[l][:, j].reshape(1, -1).T) / nl[l]
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(v[j, np.arange(L)!=l])
            if pen2 == "sign":
                s2 = mu2 * np.sum([x/np.sqrt(x**2 + 0.5) for x in v[j, np.arange(L)!=l]]) / np.sqrt(v[j, l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = [fun_s(j, l) for j in range(q) for l in range(L)]
        s = np.array(result_s).reshape(q, L)
        def fun_c(j, l):
            if pen2 == "magnitude":
                c = np.sign(s[j, l]) * (abs(s[j, l]) > mu[j, l]) * (abs(s[j, l]) - mu[j, l]) / (mu2 * (L - 1))
            if pen2 == "sign":
                c = np.sign(s[j, l]) * (abs(s[j, l]) > mu[j, l]) * (abs(s[j, l]) - mu[j, l]) / (mu2 * (L - 1) / (u[j, l]**2 + 0.5))
            return c
        c = np.array([fun_c(j, l) for j in range(q) for l in range(L)]).reshape(q, L)
        return c
    what = np.zeros((p, L))
    def fun_1(l):
        Z_l = svds(np.dot(x[l].T, y[l]), k=1)
        u_l = Z_l[0]
        return u_l
    U = np.hstack(np.array([fun_1(l) for l in range(L)]))
    def fun_2(l):
        Z_l = svds(np.dot(x[l].T, y[l]), k=1)
        v_l = Z_l[2]
        return v_l
    V = np.vstack(np.array([fun_2(l) for l in range(L)])).T
    sgn1 = [np.sign(np.dot(np.dot(U[:, l], x[l].T), np.dot(y[l], V[:, l]))) for l in range(L)]
    for l in range(L):
        V[:, l] = sgn1[l] * V[:, l]
    sgn2 = [np.sign(np.dot(U[:, 0], U[:, l]) / (np.sqrt(np.sum(U[:, 0]**2)) * np.sqrt(np.sum(U[:, l]**2)))) for l in range(1, L)]
    for l in range(1, L):
        U[:, l] = sgn2[l - 1] * U[:, l]
        V[:, l] = sgn2[l - 1] * V[:, l]
    u = U.copy()
    v = V.copy()
    iter = 1
    dis_u_iter = 10
    dis_v_iter = 10
    loading_trace_u = np.zeros((p * L, maxstep))
    loading_trace_v = np.zeros((q * L, maxstep))
    if trace:
        print("The variables that join the set of selected variables at each step:")
    while dis_u_iter > eps and dis_v_iter > eps and iter <= maxstep:
        u_iter = u.copy()
        v_iter = v.copy()
        subiter_u = 1
        dis_u = 10
        while dis_u > eps and subiter_u <= submaxstep:
            u_old = u.copy()
            if pen1 == "homogeneity":
                u = u_value_homo(u, v, p, q, L, mu1, mu2, pen2)
            if pen1 == "heterogeneity":
                u = u_value_hetero(u, v, p, q, L, mu1, mu2, pen2)
            u_norm = np.sqrt(np.sum(u**2, axis=0))
            u_norm[u_norm == 0] = 1e-04
            u = (u.T / np.expand_dims(u_norm, axis=-1)).T
            dis_u = np.max(np.sqrt(np.sum((u - u_old)**2, axis=0)) / np.sqrt(np.sum(u_old**2, axis=0)))
            what = u
            what_cut = np.where(np.abs(what) > 1e-04, what, 0)
            what_cut_norm = np.sqrt(np.sum(what_cut**2, axis=0))
            what_cut_norm[what_cut_norm == 0] = 1e-04
            what_dir = (what_cut.T / np.expand_dims(what_cut_norm, axis=-1)).T
            what_cut = np.where(np.abs(what_dir) > 1e-04, what_dir, 0)
            if np.sum(np.sum(np.abs(u), axis=0) <= 1e-04) == p:
                print("Stop! The value of mu1 is too large")
                break
            subiter_u += 1
        loading_trace_u[:, iter-1] = what_cut.ravel(order='F')
        subiter_v = 1
        dis_v = 10
        while dis_v > eps and subiter_v <= submaxstep:
            v_old = v.copy()
            if pen1 == "homogeneity":
                v = v_value_homo(u, v, p, q, L, mu1=mu3, mu2=mu4, pen2=pen2)
            if pen1 == "heterogeneity":
                v = v_value_hetero(u, v, p, q, L, mu1=mu3, mu2=mu4, pen2=pen2)
            v_norm = np.sqrt(np.sum(v**2, axis=0))
            v_norm[v_norm == 0] = 1e-04
            v = (v.T / np.expand_dims(v_norm, axis=-1)).T
            dis_v = np.max(np.sqrt(np.sum((v - v_old)**2, axis=0)) / np.sqrt(np.sum(v_old**2, axis=0)))
            what_v = v
            what_cut_v = np.where(np.abs(what_v) > 1e-04, what_v, 0)
            what_cut_norm = np.sqrt(np.sum(what_cut_v**2, axis=0))
            what_cut_norm[what_cut_norm == 0] = 1e-04
            what_dir = (what_cut_v.T / np.expand_dims(what_cut_norm, axis=-1)).T
            what_cut_v = np.where(np.abs(what_dir) > 1e-04, what_dir, 0)
            if np.sum(np.sum(np.abs(v), axis=0) <= 1e-04) == q:
                print("Stop! The value of mu1 is too large")
                break
            subiter_v += 1
        loading_trace_v[:, iter-1] = what_cut_v.ravel(order='F')
        dis_u_iter = np.max(np.sqrt(np.sum((u - u_iter)**2, axis=0)) / np.sqrt(np.sum(u_iter**2, axis=0)))
        dis_v_iter = np.max(np.sqrt(np.sum((v - v_iter)**2, axis=0)) / np.sqrt(np.sum(v_iter**2, axis=0)))
        if np.sum(np.sum(np.abs(u), axis=0) <= 1e-04) == p or np.sum(np.sum(np.abs(v), axis=0) <= 1e-04) == q:
            break
        if trace:
            new2A = [np.where(what_cut[:, l] != 0)[0] for l in range(L)]
            print("\n--------------------\n----- Step {} -----\n--------------------".format(iter))
            for l in range(L):
                new2A_l = new2A[l]
                if len(new2A_l) <= 10:
                    print("DataSet {}:".format(l+1))
                    print(', '.join(['X{}'.format(i+1) for i in new2A_l]))
                else:
                    print("DataSet {}:".format(l+1))
                    nlines = int(np.ceil(len(new2A_l) / 10))
                    for i in range(nlines - 1):
                        print(', '.join(['X{}'.format(i+1) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                    print(', '.join(['X{}'.format(i+1) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
            new2A_v = [np.where(what_cut_v[:, l] != 0)[0] for l in range(L)]
            for l in range(L):
                new2A_l = new2A_v[l]
                if len(new2A_l) <= 10:
                    print("DataSet {}:".format(l+1))
                    print(', '.join(['Y{}'.format(i+1) for i in new2A_l]))
                else:
                    print("DataSet {}:".format(l+1))
                    nlines = int(np.ceil(len(new2A_l) / 10))
                    for i in range(nlines - 1):
                        print(', '.join(['Y{}'.format(i+1) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                    print(', '.join(['Y{}'.format(i+1) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
        iter += 1
    if iter > 1:
        loading_trace_u = loading_trace_u[:, :iter-1]
        loading_trace_v = loading_trace_v[:, :iter-1]
    what_u = what_cut
    what_v = what_cut_v
    new2A = [ip[what_u[:, l] != 0] for l in range(L)]
    new2A_v = [iq[what_v[:, l] != 0] for l in range(L)]
    def plot_draw(draw):
        if draw:
            for l in range(L):
                plt.figure(figsize=(10, 5))
                plt.subplot(1, 2, 1)
                plt.plot(np.transpose(loading_trace_u[l * p : (l + 1) * p, ]))
                plt.title(f'Dataset {l+1}\nConvergence path of elements in vector u')
                plt.xlabel('Number of iterations')
                plt.ylabel('Weight')
                plt.subplot(1, 2, 2)
                plt.plot(np.transpose(loading_trace_v[l * q : (l + 1) * q, ]))
                plt.title(f'Dataset {l+1}\nConvergence path of elements in vector v')
                plt.xlabel('Number of iterations')
                plt.ylabel('Weight')
                plt.tight_layout()
                plt.show()
    plot_draw(draw = draw)  
    dictname = [f'Dataset {l+1}' for l in range(L)]
    new2A_dict = {}
    new2A_v_dict = {}
    meanx_dict = {}
    meany_dict = {}
    normx_dict = {}
    normy_dict = {}
    x_dict = {}
    y_dict = {}    
    for l in range(L):
        new2A_dict.update({dictname[l]:new2A[l]})
        new2A_v_dict.update({dictname[l]:new2A_v[l]})
        meanx_dict.update({dictname[l]:meanx[l]})
        meany_dict.update({dictname[l]:meany[l]})
        normx_dict.update({dictname[l]:normx[l]})
        normy_dict.update({dictname[l]:normy[l]})        
        x_dict.update({dictname[l]:x[l]})
        y_dict.update({dictname[l]:y[l]}) 
    what_df = pd.DataFrame(what, index = [str(i+1) for i in range(p)], columns = dictname)
    what_v_df = pd.DataFrame(what_v, index = [str(i+1) for i in range(p)], columns = dictname)
    class iscca:
        def __init__(self, data):
            self.x = data['x']
            self.y = data['y']
            self.loading_x = data['loading.x']
            self.loading_y = data['loading.y']
            self.variable_x = data['variable.x']
            self.variable_y = data['variable.y']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.meany = data['meany']
            self.normy = data['normy']
            self.pen1 = data['pen1']
            self.pen2 = data['pen2']
            self.mu1 = data['mu1']
            self.mu2 = data['mu2']
            self.mu3 = data['mu3']
            self.mu4 = data['mu4']
            self.loading_trace_u = data['loading_trace_u']
            self.loading_trace_v = data['loading_trace_v']
    object_dict = {
        'x': x_dict, 
        'y': y_dict, 
        'loading.x': what_df, 
        'loading.y': what_v_df, 
        'variable.x': new2A_dict, 
        'variable.y': new2A_v_dict, 
        'meanx': meanx_dict, 
        'normx': normx_dict, 
        'meany': meany_dict, 
        'normy': normy_dict, 
        'pen1': pen1, 
        'pen2': pen2, 
        'mu1': mu1, 
        'mu2': mu2, 
        'mu3': mu3, 
        'mu4': mu4, 
        'loading_trace_u': loading_trace_u, 
        'loading_trace_v': loading_trace_v
    }
    object = iscca(object_dict)
    return object


import numpy as np
from scipy.integrate import quad
from sklearn.preprocessing import scale
from scipy.sparse.linalg import svds

def spca(x, mu1, eps=1e-04, scale_x=True, maxstep=50, trace=False):
    x = np.matrix(x)
    n, p = x.shape
    ip = np.arange(1,p+1)
    one = np.ones((1, n))
    meanx = one.dot(x) / n
    x = scale(x, with_std=False)
    if scale_x:
        normx = np.sqrt(np.dot(one, x**2) / (n - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        x = scale(x, with_std=False) / normx
    else:
        normx = np.ones(p)
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    what = np.zeros((p, 1))
    u, s, vt = svds(x, k=1)
    U = vt.T * s[0]
    V = u
    u = U
    v = V
    iter = 1
    dis_u = 10
    loading_trace = np.zeros((p, maxstep))
    if trace:
        print("The variables that join the set of selected variables at each step:")
    while dis_u > eps and iter <= maxstep:
        u_old = u.copy()
        v_old = v.copy()
        s = np.sum(x * np.tile(v, p).reshape(n, p), axis=0) / n
        ro_d = [ro_d1st(s[j], mu1, 6) for j in range(p)]
        fun_c = lambda j: np.sign(s[j]) * (abs(s[j]) > ro_d[j]) * (abs(s[j]) - ro_d[j])
        u = np.array([fun_c(j) for j in range(p)])
        if np.sum(np.abs(u) <= 1e-04) == p:
            print("The value of mu1 is too large")
            what_cut = u
            break
        v = np.expand_dims(x.dot(u) / np.sqrt(np.sum((x.dot(u))**2)),axis=-1)
        u_norm = np.sqrt(np.sum(u**2))
        u_norm = 1e-04 if u_norm == 0 else u_norm
        u_scale = u / u_norm
        dis_u = np.sqrt(np.sum((u - u_old)**2)) / np.sqrt(np.sum(u_old**2))
        what = u_scale
        what_cut = np.where(np.abs(what) > 1e-04, what, 0)
        what_cut_norm = np.sqrt(np.sum(what_cut**2))
        what_cut_norm = 1e-04 if what_cut_norm == 0 else what_cut_norm
        what_dir = what_cut / what_cut_norm
        what_cut = np.where(np.abs(what_dir) > 1e-04, what_dir, 0)
        loading_trace[:, iter-1] = what_cut
        if np.sum(np.abs(u_scale) <= 1e-04) == p:
            print("The value of mu1 is too large")
            break
        if trace:
            new2A = ip[what_cut != 0]
            print("\n--------------------\n----- Step {} -----\n--------------------".format(iter))
            new2A_l = new2A
            if len(new2A_l) <= 10:
                print(', '.join(['X{}'.format(i) for i in new2A_l]))
            else:
                nlines = int(np.ceil(len(new2A_l) / 10))
                for i in range(nlines - 1):
                    print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
        iter += 1
    loading_trace = loading_trace[:, :iter-2]
    what = what_cut
    new2A = ip[what != 0]
    eigenvalue = what.T.dot(np.cov(x, rowvar = False)).dot(what)
    comp = x.dot(what)
    class spca:
        def __init__(self, data):
            self.x = data['x']
            self.eigenvalue = data['eigenvalue']
            self.eigenvector = data['eigenvector']
            self.component = data['component']
            self.variable = data['variable']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.mu1 = data['mu1']
    object_dict = {
        'x': x, 
        'eigenvalue': eigenvalue, 
        'eigenvector': what, 
        'component': comp, 
        'variable': new2A, 
        'meanx': meanx,
        'normx': normx, 
        'mu1': mu1, 
    }
    object = spca(object_dict)
    return object


import os
import pickle
import pkg_resources

def cca_data():
    simData_cca_path = pkg_resources.resource_filename(__name__, 'simData_cca.py')
    script_dir = os.path.dirname(simData_cca_path)
    pkl_file_path = os.path.join(script_dir, 'cca_dict.pkl')
    with open(pkl_file_path, 'rb') as f:
        cca_dict = pickle.load(f)
    return cca_dict



import numpy as np
from scipy.integrate import quad
from sklearn.preprocessing import scale
from scipy.sparse.linalg import svds
from sklearn.model_selection import KFold
from numpy.linalg import norm
import itertools
from .Ispca import ispca

def ispca_cv(x, L, mu1, mu2, K = 5, eps = 1e-04, pen1 = "homogeneity", 
             pen2 = "magnitude", scale_x = True, maxstep = 50, submaxstep = 10):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    x = [np.array(i) for i in x] 
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x] 
    p = list(set(pl))
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    ip = np.array(range(1, p+1))
    n = np.mean(np.unique(nl)) 
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    folds = [list(KFold(n_splits=K, shuffle = True).split(range(nl[l]))) for l in range(L)]
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    def u_value_homo(u, v, p, L, mu1, mu2, pen2, x, nl):
        def fun_s(j, l):
            s1 = np.sum(x[l] * np.tile(v[l], (p)).reshape(nl[l], p), axis=0) / nl[l]
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(u[j, np.arange(len(u[0]))!=l])
            if pen2 == "sign":
                s2 = mu2 * np.sum([x / np.sqrt(x**2 + 0.5) for x in u[j, np.arange(len(u[0]))!=l]]) / np.sqrt(u[j, l]**2 + 0.5)
            s = s1[j] + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)])
        s = result_s.reshape(p, L)
        norm_u_j = np.apply_along_axis(lambda x: np.sqrt(np.sum(x**2)), 1, u)
        ro_d = ro_d1st(norm_u_j, mu1, 6)
        def fun_c(j, l):
            s_norm = np.sqrt(np.sum(s[j, :]**2))
            if pen2 == "magnitude":
                c = nl[l] * (s_norm > ro_d[j]) * s[j, l] * (s_norm - ro_d[j]) / ((1 + mu2 * nl[l] * (L - 1)) * s_norm)
            if pen2 == "sign":
                c = nl[l] * (s_norm > ro_d[j]) * s[j, l] * (s_norm - ro_d[j]) / ((1 + mu2 * nl[l] * (L - 1) / (u[j, l]**2 + 0.5)) * s_norm)
            return c
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def u_value_hetero(u, v, p, L, mu1, mu2, pen2, x, nl):
        def fun_mu(j, l):
            u_j = u[j, :]
            ro_j = ro(np.abs(u_j), mu1, 6)
            s_ro = np.sum(ro_j)
            mu_jl = ro_d1st(s_ro, 1, 1/2 * L * 6 * mu1**2) * ro_d1st(abs(u_j[l]), mu1, 6)
            return mu_jl
        result_mu = np.array([fun_mu(j, l) for j in range(p) for l in range(L)])
        mu = result_mu.reshape(p, L)
        def fun_s(j, l):
            s1 = np.sum(x[l] * np.tile(v[l], (p)).reshape(nl[l], p), axis=0) / nl[l]
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(u[j, np.arange(len(u[0]))!=l])
            if pen2 == "sign":
                s2 = mu2 * np.sum([x / np.sqrt(x**2 + 0.5) for x in u[j, np.arange(len(u[0]))!=l]]) / np.sqrt(u[j, l]**2 + 0.5)
            s = s1[j] + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)])
        s = result_s.reshape(p, L)
        def fun_c(j, l):
            if pen2 == "magnitude":
                c = nl[l] * np.sign(s[j, l]) * (abs(s[j, l]) > mu[j, l]) * (abs(s[j, l]) - mu[j, l]) / (1 + mu2 * nl[l] * (L - 1))
            if pen2 == "sign":
                c = nl[l] * np.sign(s[j, l]) * (abs(s[j, l]) > mu[j, l]) * (abs(s[j, l]) - mu[j, l]) / (1 + mu2 * nl[l] * (L - 1) / (u[j, l]**2 + 0.5))
            return c   
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def fun_k(k):
        x_train = [x[l][folds[l][k][0]] for l in range(L)]
        x_test = [x[l][folds[l][k][1]] for l in range(L)]
        nl_train = [len(x_t) for x_t in x_train]
        nl_test = [len(x_t) for x_t in x_test]
        return {'x_train': x_train, 'x_test': x_test, 'nl_train': nl_train, 'nl_test': nl_test}
    data_cv = [fun_k(k) for k in range(K)]
    def cv_fun(k):
        x_train = data_cv[k]['x_train']
        x_test = data_cv[k]['x_test']
        nl_train = data_cv[k]['nl_train']
        nl_test = data_cv[k]['nl_test']
        what = np.zeros((p, L))
        def fun_1(l):
             u_l, s_l, v_l = svds(x_train[l], k=1)
             return v_l.T * s_l[0]
        U = np.hstack(np.array([fun_1(l) for l in range(L)]))
        def fun_2(l):
            u_l, s_l, v_l = svds(x_train[l], k=1)
            return u_l
        V = [fun_2(l) for l in range(L)]
        sgn = [np.sign(np.dot(U[:, 0], U[:, l]) / (np.sqrt(np.sum(U[:, 0]**2)) * np.sqrt(np.sum(U[:, l]**2)))) for l in range(1, L)]
        for l in range(1, L):
            U[:, l] = sgn[l - 1] * U[:, l]
            V[l] = sgn[l - 1] * V[l]
        u = U.copy()
        v = V.copy()
        iter = 1
        dis_u_iter = 10
        while dis_u_iter > eps and iter <= maxstep:
            u_iter = u.copy()
            subiter_u = 1
            dis_u = 10
            while dis_u > eps and subiter_u <= submaxstep:
                u_old = u.copy()
                if pen1 == "homogeneity":
                    u = u_value_homo(u, v, p, L, mu1, mu2, pen2, x = x_train, nl = nl_train)
                if pen1 == "heterogeneity":
                    u = u_value_hetero(u, v, p, L, mu1, mu2, pen2, x = x_train, nl = nl_train)
                u_norm = np.sqrt(np.sum(u**2, axis=0))
                u_norm[u_norm == 0] = 1e-04
                u_scale = (u.T / np.expand_dims(u_norm, axis=-1)).T  
                dis_u = np.max(np.sqrt(np.sum((u - u_old)**2, axis=0)) / np.sqrt(np.sum(u_old**2, axis=0)))
                what = u_scale
                what_cut = np.where(np.abs(what) > 1e-04, what, 0)
                what_cut_norm = np.sqrt(np.sum(what_cut**2, axis=0))
                what_cut_norm[what_cut_norm == 0] = 1e-04
                what_dir = (what_cut.T / np.expand_dims(what_cut_norm, axis=-1)).T
                what_cut = np.where(np.abs(what_dir) > 1e-04, what_dir, 0)
                if np.sum(np.sum(np.abs(u_scale), axis=0) <= 1e-04) == p:
                    print("The value of mu1 is too large")
                    break
                subiter_u += 1
            if np.sum(np.sum(np.abs(u), axis=0) <= 1e-04) == p:
                print("The value of mu1 is too large")
                break  
            v = [np.expand_dims(x_train[l].dot(u[:, l]) / np.sqrt(np.sum((x_train[l].dot(u[:, l])**2))), axis=-1) for l in range(L)]
            dis_u_iter = np.max(np.sqrt(np.sum((u - u_iter)**2, axis=0)) / np.sqrt(np.sum(u_iter**2, axis=0)))
            if np.sum(np.sum(np.abs(u), axis=0) <= 1e-04) == p:
                break
            iter += 1
        v_test = [np.expand_dims(np.dot(x_test[l], what_cut[:, l]),axis=-1) for l in range(L)]
        def fun_3(l):
            rho = norm(x_test[l].T - np.dot(np.expand_dims(what_cut[:, l],axis=-1), v_test[l].T), 'fro')**2 / nl_test[l]
            return rho
        cv_rho = sum([fun_3(l) for l in range(L)])
        return cv_rho  
    mu_list = [mu1, mu2]
    for i in range(len(mu_list)):
        if not isinstance(mu_list[i], list):
            mu_list[i] = [mu_list[i]]
    mu1, mu2 = mu_list
    mu = np.array(list(itertools.product(mu1, mu2)))
    for i in range(len(mu_list)):
        if isinstance(mu_list[i], list) and len(mu_list[i]) > 1:
            mu_list[i] = np.array(mu_list[i])
        else:
            mu_list[i] = float(mu_list[i][0])
    mu1, mu2 = mu_list
    rho = []
    for loop in range(mu.shape[0]):
        mu1 = mu[loop, 0]
        mu2 = mu[loop, 1]
        cv_fun_list = []
        for k in range(K):
            cv_fun_list.append(cv_fun(k))
        rho.append(np.mean(cv_fun_list))
    index = np.argmin(rho)
    mu1_final = mu[index, 0]
    mu2_final = mu[index, 1]
    result = ispca(x, L, mu1 = mu1_final, mu2 = mu2_final, eps=eps, pen1 = pen1, 
                    pen2 = pen2, scale_x = scale_x, maxstep = maxstep, trace = False, draw = False)
    dictname = [f'Dataset {l+1}' for l in range(L)]
    folds_dict = {}
    meanx_dict = {}
    normx_dict = {}
    x_dict = {}   
    for l in range(L):
        folds_dict.update({dictname[l]:folds[l]})
        meanx_dict.update({dictname[l]:meanx[l]})
        normx_dict.update({dictname[l]:normx[l]})       
        x_dict.update({dictname[l]:x[l]})           
    class ispca_cv:
        def __init__(self, data):
            self.x = data['x']
            self.mu1 = data['mu1']
            self.mu2 = data['mu2']
            self.fold = data['fold']
            self.eigenvalue = data['eigenvalue']
            self.eigenvector = data['eigenvector']
            self.component = data['component']
            self.variable = data['variable']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.pen1 = data['pen1']
            self.pen2 = data['pen2']
            self.loading_trace = data['loading_trace']
    object_dict = {
        'x': x_dict, 
        'mu1': mu1_final, 
        'mu2': mu2_final, 
        'fold': folds_dict,
        'eigenvalue': result.eigenvalue,
        'eigenvector': result.eigenvector,
        'component': result.component, 
        'variable': result.variable, 
        'meanx': meanx_dict, 
        'normx': normx_dict, 
        'pen1': pen1, 
        'pen2': pen2, 
        'loading_trace': result.loading_trace, 
    }
    object = ispca_cv(object_dict)
    return object


import numpy as np
import pandas as pd
from sklearn.preprocessing import scale
from scipy.sparse.linalg import svds
import matplotlib.pyplot as plt
import seaborn as sns

def preview_pca(x, L, scale_x=True):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    x = [np.array(i) for i in x]
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x]     
    p = list(set(pl)) 
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    what = np.zeros((p, L))
    def fun_1(l):
         u_l, s_l, v_l = svds(x[l], k=1)
         return v_l.T
    U = np.hstack(np.array([fun_1(l) for l in range(L)]))
    what = U
    eigenvalue = [what[:, l].T.dot(np.cov(x[l], rowvar = False)).dot(what[:, l]) for l in range(L)]
    comp = [x[l].dot(what[:, l]) for l in range(L)]
    dictname = [f'Dataset {l+1}' for l in range(L)]
    meanx_dict = {}
    normx_dict = {}
    x_dict = {}
    eigenvalue_dict = {}
    comp_dict = {}
    for l in range(L):
        meanx_dict.update({dictname[l]:meanx[l]})
        normx_dict.update({dictname[l]:normx[l]})      
        x_dict.update({dictname[l]:x[l]})
        eigenvalue_dict.update({dictname[l]:eigenvalue[l]})
        comp_dict.update({dictname[l]:comp[l]})
    what_df = pd.DataFrame(what, index = [str(i+1) for i in range(p)], columns = dictname)
    dictname = [f'Dataset {l+1}' for l in range(L)]
    meanx_dict = {}
    normx_dict = {}
    x_dict = {}   
    for l in range(L):
        meanx_dict.update({dictname[l]:meanx[l]})
        normx_dict.update({dictname[l]:normx[l]})       
        x_dict.update({dictname[l]:x[l]})
    what_df = pd.DataFrame(what, index = [str(i+1) for i in range(p)], columns = dictname)
    for l in range(L):
        plt.figure()
        plt.scatter(x = np.arange(1, p+1), y = what_df.iloc[:, l])
        plt.title(f"The first principal component of dataset {l+1}")
        plt.xlabel("Dimension")
        plt.ylabel("Value")
    heat = [np.cov(i) for i in x]
    for l in range(L):
        plt.figure()
        sns.heatmap(heat[l], cmap = 'Reds', xticklabels=False, yticklabels=False)
        plt.xlabel(f"Dataset {l+1}: X")
        plt.ylabel("Y")
        plt.title("Heatmap of covariance")
    class preview_pca:
        def __init__(self, data):
            self.x = data['x']
            self.eigenvalue = data['eigenvalue']
            self.eigenvector = data['eigenvector']
            self.component = data['component']
            self.meanx = data['meanx']
            self.normx = data['normx']
    object_dict = {
        'x': x_dict, 
        'eigenvalue': eigenvalue_dict, 
        'eigenvector': what_df, 
        'component': comp_dict, 
        'meanx': meanx_dict,
        'normx': normx_dict, 
    }
    object = preview_pca(object_dict)
    return object


import numpy as np
import pandas as pd
from scipy.integrate import quad
from sklearn.preprocessing import scale
from scipy.sparse.linalg import svds
import matplotlib.pyplot as plt

def ispca(x, L, mu1, mu2, eps = 1e-04, pen1 = "homogeneity", 
          pen2 = "magnitude", scale_x = True, maxstep = 50, 
          submaxstep = 10, trace = False, draw = False):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    x = [np.array(i) for i in x] 
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x] 
    p = list(set(pl))
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    ip = np.array(range(1, p+1))
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    def u_value_homo(u, v, p, L, mu1, mu2, pen2):
        def fun_s(j, l):
            s1 = np.sum(x[l] * np.tile(v[l], (p)).reshape(nl[l], p), axis=0) / nl[l]
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(u[j, np.arange(len(u[0]))!=l])
            if pen2 == "sign":
                s2 = mu2 * np.sum([x / np.sqrt(x**2 + 0.5) for x in u[j, np.arange(len(u[0]))!=l]]) / np.sqrt(u[j, l]**2 + 0.5)
            s = s1[j] + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)])
        s = result_s.reshape(p, L)
        norm_u_j = np.apply_along_axis(lambda x: np.sqrt(np.sum(x**2)), 1, u)
        ro_d = ro_d1st(norm_u_j, mu1, 6)
        def fun_c(j, l):
            s_norm = np.sqrt(np.sum(s[j, :]**2))
            if pen2 == "magnitude":
                c = nl[l] * (s_norm > ro_d[j]) * s[j, l] * (s_norm - ro_d[j]) / ((1 + mu2 * nl[l] * (L - 1)) * s_norm)
            if pen2 == "sign":
                c = nl[l] * (s_norm > ro_d[j]) * s[j, l] * (s_norm - ro_d[j]) / ((1 + mu2 * nl[l] * (L - 1) / (u[j, l]**2 + 0.5)) * s_norm)
            return c
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def u_value_hetero(u, v, p, L, mu1, mu2, pen2):
        def fun_mu(j, l):
            u_j = u[j, :]
            ro_j = ro(np.abs(u_j), mu1, 6)
            s_ro = np.sum(ro_j)
            mu_jl = ro_d1st(s_ro, 1, 1/2 * L * 6 * mu1**2) * ro_d1st(abs(u_j[l]), mu1, 6)
            return mu_jl
        result_mu = np.array([fun_mu(j, l) for j in range(p) for l in range(L)])
        mu = result_mu.reshape(p, L)
        def fun_s(j, l):
            s1 = np.sum(x[l] * np.tile(v[l], (p)).reshape(nl[l], p), axis=0) / nl[l]
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(u[j, np.arange(len(u[0]))!=l])
            if pen2 == "sign":
                s2 = mu2 * np.sum([x / np.sqrt(x**2 + 0.5) for x in u[j, np.arange(len(u[0]))!=l]]) / np.sqrt(u[j, l]**2 + 0.5)
            s = s1[j] + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)])
        s = result_s.reshape(p, L)
        def fun_c(j, l):
            if pen2 == "magnitude":
                c = nl[l] * np.sign(s[j, l]) * (abs(s[j, l]) > mu[j, l]) * (abs(s[j, l]) - mu[j, l]) / (1 + mu2 * nl[l] * (L - 1))
            if pen2 == "sign":
                c = nl[l] * np.sign(s[j, l]) * (abs(s[j, l]) > mu[j, l]) * (abs(s[j, l]) - mu[j, l]) / (1 + mu2 * nl[l] * (L - 1) / (u[j, l]**2 + 0.5))
            return c   
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    what = np.zeros((p, L))
    def fun_1(l):
        u_l, s_l, v_l = svds(x[l], k=1)
        return v_l.T * s_l[0]
    U = np.hstack(np.array([fun_1(l) for l in range(L)]))
    def fun_2(l):
        u_l, s_l, v_l = svds(x[l], k=1)
        return u_l
    V = [fun_2(l) for l in range(L)]
    sgn = [np.sign(np.dot(U[:, 0], U[:, l]) / (np.sqrt(np.sum(U[:, 0]**2)) * np.sqrt(np.sum(U[:, l]**2)))) for l in range(1, L)]
    for l in range(1, L):
        U[:, l] = sgn[l - 1] * U[:, l]
        V[l] = sgn[l - 1] * V[l]
    u = U.copy()
    v = V.copy()
    iter = 1
    dis_u_iter = 10
    loading_trace = np.zeros((p * L, maxstep))
    if trace:
        print("The variables that join the set of selected variables at each step:")
    while dis_u_iter > eps and iter <= maxstep:
        u_iter = u.copy()
        subiter_u = 1
        dis_u = 10
        while dis_u > eps and subiter_u <= submaxstep:
            u_old = u.copy()
            if pen1 == "homogeneity":
                u = u_value_homo(u, v, p, L, mu1, mu2, pen2)
            if pen1 == "heterogeneity":
                u = u_value_hetero(u, v, p, L, mu1, mu2, pen2)
            u_norm = np.sqrt(np.sum(u**2, axis=0))
            u_norm[u_norm == 0] = 1e-04
            u_scale = (u.T / np.expand_dims(u_norm, axis=-1)).T  
            dis_u = np.max(np.sqrt(np.sum((u - u_old)**2, axis=0)) / np.sqrt(np.sum(u_old**2, axis=0)))
            what = u_scale
            what_cut = np.where(np.abs(what) > 1e-04, what, 0)
            what_cut_norm = np.sqrt(np.sum(what_cut**2, axis=0))
            what_cut_norm[what_cut_norm == 0] = 1e-04
            what_dir = (what_cut.T / np.expand_dims(what_cut_norm, axis=-1)).T
            what_cut = np.where(np.abs(what_dir) > 1e-04, what_dir, 0)
            loading_trace[:, iter-1] = what_cut.ravel(order='F')
            if np.sum(np.sum(np.abs(u_scale), axis=0) <= 1e-04) == p:
                break
            subiter_u += 1
        if np.sum(np.sum(np.abs(u), axis=0) <= 1e-04) == p:
            print("Stop! The value of mu1 is too large")
            what_cut = u
            break  
        v = [np.expand_dims(x[l].dot(u[:, l]) / np.sqrt(np.sum((x[l].dot(u[:, l])**2))), axis=-1) for l in range(L)]
        dis_u_iter = np.max(np.sqrt(np.sum((u - u_iter)**2, axis=0)) / np.sqrt(np.sum(u_iter**2, axis=0)))
        if np.sum(np.sum(np.abs(u), axis=0) <= 1e-04) == p:
            break
        if trace:
            new2A = [np.where(what_cut[:, l] != 0)[0] for l in range(L)]
            print("\n--------------------\n----- Step {} -----\n--------------------".format(iter))
            for l in range(L):
                new2A_l = new2A[l]
                if len(new2A_l) <= 10:
                    print("DataSet {}:".format(l+1))
                    print(', '.join(['X{}'.format(i+1) for i in new2A_l]))
                else:
                    print("DataSet {}:".format(l+1))
                    nlines = int(np.ceil(len(new2A_l) / 10))
                    for i in range(nlines - 1):
                        print(', '.join(['X{}'.format(i+1) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                    print(', '.join(['X{}'.format(i+1) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
        iter += 1
    loading_trace = loading_trace[:, :iter-1]
    what = what_cut
    new2A = [ip[what[:, l] != 0] for l in range(L)] 
    if draw:
        for l in range(L):
            plt.figure()
            plt.plot(loading_trace[l * p: (l + 1) * p, :].T)
            plt.title(f"Dataset {l+1}\nConvergence path of elements in vector u")
            plt.xlabel("Number of iterations")
            plt.ylabel("Weight")
            plt.show()
    eigenvalue = [what[:, l].T.dot(np.cov(x[l], rowvar = False)).dot(what[:, l]) for l in range(L)]
    comp = [x[l].dot(what[:, l]) for l in range(L)]
    dictname = [f'Dataset {l+1}' for l in range(L)]
    new2A_dict = {}
    meanx_dict = {}
    normx_dict = {}
    x_dict = {}
    eigenvalue_dict = {}
    comp_dict = {}
    for l in range(L):
        new2A_dict.update({dictname[l]:new2A[l]})
        meanx_dict.update({dictname[l]:meanx[l]})
        normx_dict.update({dictname[l]:normx[l]})      
        x_dict.update({dictname[l]:x[l]})
        eigenvalue_dict.update({dictname[l]:eigenvalue[l]})
        comp_dict.update({dictname[l]:comp[l]})
    what_df = pd.DataFrame(what, index = [str(i+1) for i in range(p)], columns = dictname)
    class ispca:
        def __init__(self, data):
            self.x = data['x']
            self.eigenvalue = data['eigenvalue']
            self.eigenvector = data['eigenvector']
            self.component = data['component']
            self.variable = data['variable']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.pen1 = data['pen1']
            self.pen2 = data['pen2']
            self.mu1 = data['mu1']
            self.mu2 = data['mu2']
            self.loading_trace = data['loading_trace']
    object_dict = {
        'x': x_dict, 
        'eigenvalue': eigenvalue_dict, 
        'eigenvector': what_df, 
        'component': comp_dict, 
        'variable': new2A_dict, 
        'meanx': meanx_dict,
        'normx': normx_dict, 
        'pen1': pen1, 
        'pen2': pen2, 
        'mu1': mu1, 
        'mu2': mu2, 
        'loading_trace': loading_trace
    }
    object = ispca(object_dict)
    return object


import numpy as np
import pandas as pd
from scipy.integrate import quad
from sklearn.preprocessing import scale
from numpy.linalg import svd
from numpy.linalg import solve
from scipy.optimize import root_scalar
from numpy.linalg import lstsq
import matplotlib.pyplot as plt

def ispls(x, y, L, mu1, mu2, eps=1e-04, kappa=0.05, pen1="homogeneity", 
          pen2="magnitude", scale_x=True, scale_y=True, maxstep=50, 
          submaxstep=10, trace=False, draw=False):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    if not isinstance(y, list):
        raise TypeError("y should be of list type.")
    x = [np.array(i) for i in x]
    y = [np.array(i) for i in y]
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x]
    ql = [i.shape[1] for i in y]
    p = list(set(pl))
    q = list(set(ql))   
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(q) > 1:
        raise ValueError("The dimension of data y should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    if len(q) == 1:
        q = q[0] 
    ip = np.array(range(1, p+1))
    iq = np.array(range(1, q+1))
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    meany = [np.dot(np.ones((1, nl[l])), y[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
        y[l] = scale(y[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    def y_scale(l):
        one = np.ones((1, nl[l]))
        normy = np.sqrt(np.dot(one, y[l]**2) / (nl[l] - 1))
        if np.any(normy < np.finfo(float).eps):
            raise ValueError("Some of the columns of the response matrix have zero variance.")
        return normy
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_y:
        normy = [y_scale(i) for i in range(L)]
    else:
        normy = [[1]*q[0]]*L
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    if scale_y:
        y = [(y[i] - np.mean(y[i], axis=0)) / normy[i] for i in range(L)]
    def fun_1(l):
        z_l = np.dot(np.transpose(x[l]), y[l])
        z_l = z_l / nl[l]
        return z_l
    z = np.hstack(np.array([fun_1(l) for l in range(L)]))
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    def c_value_homo(z, a, c, p, q, L, mu1, mu2, pen2):
        def fun_s(j, l):
            z_l = z[:, (l * q):(l + 1) * q]
            a_l = a[:, l].reshape(-1, 1)
            c_j = c[j, :]
            s1 = np.dot(a_l.T, np.dot(z_l, z_l[j, :].reshape(1, -1).T)) / (nl[l]**2)
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(c_j[np.arange(len(c_j))!=l])
            elif pen2 == "sign":
                s2 = mu2 * np.sum(np.array([x / np.sqrt(x**2 + 0.5) for x in c_j[np.arange(len(c_j))!=l]])) / np.sqrt(c_j[l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        norm_c_j = np.sqrt(np.sum(c**2, axis=1))
        ro_d = ro_d1st(norm_c_j, mu1, 6)
        def fun_c(j, l):
            s_norm = np.sqrt(np.sum(result_s[j, :]**2))
            if pen2 == "magnitude":
                temp = (s_norm > ro_d[j]) * result_s[j, l] * (s_norm - ro_d[j]) / ((1/q/(nl[l]**2) + mu2 * (L - 1)) * s_norm)
            elif pen2 == "sign":
                temp = (s_norm > ro_d[j]) * result_s[j, l] * (s_norm - ro_d[j]) / ((1/q/(nl[l]**2) + mu2 * (L - 1)/(c[j, l]**2 + 0.5)) * s_norm)
            return temp
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def c_value_hetero(z, a, c, p, q, L, mu1, mu2, pen2):
        def fun_mu(j, l):
            c_j = c[j, :]
            ro_j = ro(np.abs(c_j), mu1, 6)
            s_ro = np.sum(ro_j)
            mu_jl = ro_d1st(s_ro, 1, 0.5 * L * 6 * mu1**2) * ro_d1st(np.abs(c_j[l]), mu1, 6)
            return mu_jl
        result_mu = np.array([fun_mu(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        def fun_s(j, l):
            z_l = z[:, (l * q):(l + 1) * q]
            a_l = a[:, l].reshape(-1, 1)
            c_j = c[j, :]
            s1 = np.dot(a_l.T, np.dot(z_l, z_l[j, :].reshape(1, -1).T)) / (nl[l]**2)
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(c_j[np.arange(len(c_j))!=l])
            elif pen2 == "sign":
                s2 = mu2 * np.sum(np.array([x / np.sqrt(x**2 + 0.5) for x in c_j[np.arange(len(c_j))!=l]])) / np.sqrt(c_j[l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        def fun_c(j, l):
            if pen2 == "magnitude":
                temp = np.sign(result_s[j, l]) * (np.abs(result_s[j, l]) > result_mu[j, l]) * (np.abs(result_s[j, l]) - result_mu[j, l]) / (1/q/(nl[l]**2) + mu2 * (L - 1))
            elif pen2 == "sign":
                temp = np.sign(result_s[j, l]) * (np.abs(result_s[j, l]) > result_mu[j, l]) * (np.abs(result_s[j, l]) - result_mu[j, l]) / (1/q/(nl[l]**2) + mu2 * (L - 1)/(c[j, l]**2 + 0.5))
            return temp
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def c_value_homo_q1(z, a, c, p, q, L, mu1, mu2, pen2):
        def fun_s(j, l):
            z_l = z[:, (l * q):(l + 1) * q]
            a_l = a[:, l].reshape(-1, 1)
            c_j = c[j, :]
            s1 = np.dot(a_l.T, np.dot(z_l, z_l[j].reshape(1, -1).T)) / (nl[l]**2)
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(c_j[np.arange(len(c_j))!=l])
            elif pen2 == "sign":
                s2 = mu2 * np.sum(np.array([x / np.sqrt(x**2 + 0.5) for x in c_j[np.arange(len(c_j))!=l]])) / np.sqrt(c_j[l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        norm_c_j = np.sqrt(np.sum(c**2, axis=1))
        ro_d = ro_d1st(norm_c_j, mu1, 6)
        def fun_c(j, l):
            s_norm = np.sqrt(np.sum(result_s[j, :]**2))
            if pen2 == "magnitude":
                temp = (s_norm > ro_d[j]) * result_s[j, l] * (s_norm - ro_d[j]) / ((1/q/(nl[l]**2) + mu2 * (L - 1)) * s_norm)
            elif pen2 == "sign":
                temp = (s_norm > ro_d[j]) * result_s[j, l] * (s_norm - ro_d[j]) / ((1/q/(nl[l]**2) + mu2 * (L - 1)/(c[j, l]**2 + 0.5)) * s_norm)
            return temp
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def c_value_hetero_q1(z, a, c, p, q, L, mu1, mu2, pen2):
        def fun_mu(j, l):
            c_j = c[j, :]
            ro_j = ro(np.abs(c_j), mu1, 6)
            s_ro = np.sum(ro_j)
            mu_jl = ro_d1st(s_ro, 1, 0.5 * L * 6 * mu1**2) * ro_d1st(np.abs(c_j[l]), mu1, 6)
            return mu_jl
        result_mu = np.array([fun_mu(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        def fun_s(j, l):
            z_l = z[:, (l * q):(l + 1) * q]
            a_l = a[:, l].reshape(-1, 1)
            c_j = c[j, :]
            s1 = np.dot(a_l.T, np.dot(z_l, z_l[j].reshape(1, -1).T)) / (nl[l]**2)
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(c_j[np.arange(len(c_j))!=l])
            elif pen2 == "sign":
                s2 = mu2 * np.sum(np.array([x / np.sqrt(x**2 + 0.5) for x in c_j[np.arange(len(c_j))!=l]])) / np.sqrt(c_j[l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        def fun_c(j, l):
            if pen2 == "magnitude":
                temp = np.sign(result_s[j, l]) * (np.abs(result_s[j, l]) > result_mu[j, l]) * (np.abs(result_s[j, l]) - result_mu[j, l]) / (1/q/(nl[l]**2) + mu2 * (L - 1))
            elif pen2 == "sign":
                temp = np.sign(result_s[j, l]) * (np.abs(result_s[j, l]) > result_mu[j, l]) * (np.abs(result_s[j, l]) - result_mu[j, l]) / (1/q/(nl[l]**2) + mu2 * (L - 1)/(c[j, l]**2 + 0.5))
            return temp
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    what = np.zeros((p, L))
    def fun_2(l):
        z_l = z[:, (l * q):(l + 1) * q]
        m_l = np.dot(z_l, z_l.T) / q
        return m_l
    m = np.hstack(np.array([fun_2(l) for l in range(L)]))
    dis = 10
    iter = 1
    kappa2 = (1 - kappa) / (1 - 2 * kappa)
    def fun_c(l):
        return svd(np.dot(z[:, (l * q):(l + 1) * q], z[:, (l * q):(l + 1) * q].T), full_matrices=False)[0][:, 0]
    c = np.array([fun_c(l) for l in range(L)]).T
    sgn1 = np.array([np.sign(np.dot(np.dot(c[:, l].T, z[:, (l * q):(l + 1) * q]), np.dot(z[:, (l * q):(l + 1) * q].T, c[:, l]))) for l in range(L)])
    for l in range(L):
        c[:, l] = sgn1[l] * c[:, l]
    sgn2 = np.array([np.sign(np.dot(c[:, 0].T, c[:, l])) / (np.sqrt(np.sum(c[:, 0]**2)) * np.sqrt(np.sum(c[:, l]**2))) for l in range(1, L)])
    for l in range(1, L):
        c[:, l] = sgn2[l - 1] * c[:, l]
    a = c
    dis_c_iter = 10
    loading_trace = np.zeros((p * L, maxstep))
    if trace:
        print("The variables that join the set of selected variables at each step:")
    while dis_c_iter > eps and iter <= maxstep:
        c_iter = c.copy()
        a_iter = a.copy()
        def fun_3(l):
            def h(lambda_):
                alpha = solve(m[:, l * p: (l + 1) * p] + lambda_ * np.eye(p), np.dot(m[:, l * p: (l + 1) * p], c[:, l]))
                obj = np.dot(alpha.T, alpha) - 1 / kappa2**2
                return obj
            while h(1e-04) * h(1e+12) > 0:
                m[:, (l - 1) * p: l * p] *= 2
                c[:, l] *= 2
            lambdas = root_scalar(h, method='brentq', bracket=[1e-04, 1e+12]).root
            a_l = kappa2 * solve(m[:, l * p: (l + 1) * p] + lambdas * np.eye(p), np.dot(m[:, l * p: (l + 1) * p], c[:, l]))
            return a_l
        a = np.array([fun_3(l) for l in range(L)]).T
        subiter_c = 1
        dis_c = 10
        while dis_c > eps and subiter_c <= submaxstep:
            c_old = c.copy()
            if pen1 == "homogeneity" and q != 1:
                c = c_value_homo(z, a, c, p, q, L, mu1, mu2, pen2)
            elif pen1 == "homogeneity" and q == 1:
                c = c_value_homo_q1(z, a, c, p, q, L, mu1, mu2, pen2)
            elif pen1 == "heterogeneity" and q != 1:
                c = c_value_hetero(z, a, c, p, q, L, mu1, mu2, pen2)
            elif pen1 == "heterogeneity" and q == 1:
                c = c_value_hetero_q1(z, a, c, p, q, L, mu1, mu2, pen2)
            c_norm = np.sqrt(np.sum(c**2, axis=0))
            c_norm[c_norm == 0] = 1e-04
            c = (c.T / np.expand_dims(c_norm,axis=-1)).T
            dis = np.max(np.sqrt(np.sum((c - c_old)**2, axis=0)) / np.sqrt(np.sum(c_old**2, axis=0)))
            what = c.copy()
            what_cut = np.where(np.abs(what) > 1e-04, what, 0)
            what_cut_norm = np.sqrt(np.sum(what_cut**2, axis=0))
            what_cut_norm[what_cut_norm == 0] = 1e-04
            what_dir = (what_cut.T / np.expand_dims(what_cut_norm, axis=-1)).T
            what_cut = np.where(np.abs(what_dir) > 1e-04, what_dir, 0)
            loading_trace[:, iter-1] = what_cut.ravel(order='F')
            if np.sum(np.apply_along_axis(lambda x: np.sum(np.abs(x) <= 1e-04) == p, axis=0, arr=c)) > 0:
                print("Stop! The value of mu1 is too large")
                break
            subiter_c += 1
        dis_c_iter = np.max(np.sqrt(np.sum((c - c_iter)**2, axis=0)) / np.sqrt(np.sum(c_iter**2, axis=0)))
        if np.sum(np.apply_along_axis(lambda x: np.sum(np.abs(x) <= 1e-04) == p, axis=0, arr=c)) > 0:
            break
        if trace:
            new2A = [ip[what_cut[:, l] != 0] for l in range(L)]
            print("\n--------------------\n----- Step {} -----\n--------------------".format(iter))
            for l in range(L):
                new2A_l = new2A[l]
                if len(new2A_l) <= 10:
                    print("DataSet {}:".format(l+1))
                    print(', '.join(['X{}'.format(i) for i in new2A_l]))
                else:
                    print("DataSet {}:".format(l+1))
                    nlines = int(np.ceil(len(new2A_l) / 10))
                    for i in range(nlines - 1):
                        print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                    print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
        iter += 1
    loading_trace = loading_trace[:, :iter - 1]
    what = what_cut
    new2A = [ip[what[:, l] != 0] for l in range(L)]
    betahat = np.zeros((p, q * L))
    def fun_fit(l):
        x_l = x[l]
        w_l = what[:, l].reshape(-1, 1)
        t_l = np.dot(x_l, w_l)
        if np.sum(w_l == 0) != p:
            y_l = y[l]
            fit_l = lstsq(t_l, y_l, rcond=None)[0]
            betahat_l = np.dot(w_l, fit_l).reshape(p, q)
        else:
            betahat_l = np.zeros((p, q))
        if x[l].shape[1] > 1:
            betahat_l = pd.DataFrame(betahat_l, index=range(1, p + 1))
        if q > 1 and y[l].shape[1] > 1:
            betahat_l.columns = range(1, q + 1)
        return betahat_l
    betahat = [fun_fit(l) for l in range(L)]
    dictname = [f'Dataset {l+1}' for l in range(L)]
    new2A_dict = {}
    betahat_dict = {}
    meanx_dict = {}
    meany_dict = {}
    normx_dict = {}
    normy_dict = {}
    x_dict = {}
    y_dict = {}    
    for l in range(L):
        new2A_dict.update({dictname[l]:new2A[l]})
        betahat_dict.update({dictname[l]:betahat[l]})
        meanx_dict.update({dictname[l]:meanx[l]})
        meany_dict.update({dictname[l]:meany[l]})
        normx_dict.update({dictname[l]:normx[l]})
        normy_dict.update({dictname[l]:normy[l]})        
        x_dict.update({dictname[l]:x[l]})
        y_dict.update({dictname[l]:y[l]}) 
    what_df = pd.DataFrame(what, index = [str(i+1) for i in range(p)], columns = dictname)
    if draw:
        for l in range(L):
            plt.figure()
            plt.plot(loading_trace[(l * p):((l + 1) * p), :].T)
            plt.title(f"Dataset {l + 1}\nConvergence path of elements in vector w")
            plt.xlabel("Number of iterations")
            plt.ylabel("Weight")
            plt.show()
    class ispls:
        def __init__(self, data):
            self.x = data['x']
            self.y = data['y']
            self.betahat = data['betahat']
            self.loading = data['loading']
            self.variable = data['variable']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.meany = data['meany']
            self.normy = data['normy']
            self.pen1 = data['pen1']
            self.pen2 = data['pen2']
            self.mu1 = data['mu1']
            self.mu2 = data['mu2']
            self.kappa = data['kappa']
            self.loading_trace = data['loading_trace']
    object_dict = {
        'x': x_dict, 
        'y': y_dict, 
        'betahat': betahat_dict, 
        'loading': what_df, 
        'variable': new2A_dict,
        'meanx': meanx_dict, 
        'normx': normx_dict, 
        'meany': meany_dict, 
        'normy': normy_dict, 
        'pen1': pen1, 
        'pen2': pen2, 
        'mu1': mu1, 
        'mu2': mu2,  
        'kappa': kappa, 
        'loading_trace': loading_trace
    }
    object = ispls(object_dict)
    return object



import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

def ispls_plot(x, type):
    betahat = x.betahat
    L = len(betahat)
    p = betahat[list(betahat.keys())[0]].shape[0]
    q = betahat[list(betahat.keys())[0]].shape[1]
    loading_trace = x.loading_trace
    iter = loading_trace.shape[1]
    if type == "path":
        for l in range(L):
            plt.figure()
            plt.plot(np.transpose(loading_trace[(l * p):((l + 1) * p), :]))
            plt.title(f"Dataset {l+1}\nConvergence path of elements in vector w")
            plt.xlabel("Number of iterations")
            plt.ylabel("Weight")
            plt.show()
    if type == "loading":
        for l in range(L):
            plt.figure()
            plt.scatter(x = range(1, p+1), y = loading_trace[(l * p):((l + 1) * p), iter-1])
            plt.title(f"Dataset {l+1}\nThe first direction vector")
            plt.xlabel("Dimension")
            plt.ylabel("Value")
            plt.show()
    if type == "heatmap" and q != 1:
        for l in range(L):
            plt.figure()
            sns.heatmap(np.transpose(betahat[list(betahat.keys())[l]]), cmap='Reds')
            plt.title(f"Dataset {l+1}: p\nHeatmap of coefficient β[PLS]")
            plt.show()
    if type == "heatmap" and q == 1:
        betahat_matrix = np.zeros((p, L))
        for l in range(L):
            betahat_matrix[:, l] = betahat[list(betahat.keys())[l]]
        plt.figure()
        sns.heatmap(np.transpose(betahat_matrix), cmap='Reds')
        plt.title("p\nHeatmap of coefficient β[PLS]")
        plt.show()



import numpy as np
from scipy.integrate import quad
from sklearn.preprocessing import scale
from scipy.sparse.linalg import svds

def scca(x, y, mu1, mu2, eps=1e-04, scale_x=True, scale_y=True, maxstep=50, trace=False):
    x = np.asmatrix(x)
    y = np.asmatrix(y)
    nx, p = x.shape
    ny, q = y.shape
    if nx != ny:
        raise ValueError("The rows of data x and data y should be consistent.") 
    n = nx
    ip = np.arange(1,p+1)
    iq = np.arange(1,q+1)
    one = np.ones((1, n))
    meany = np.dot(one, y) / n
    y = scale(y, with_std=False)
    meanx = np.dot(one, x) / n
    x = scale(x, with_std=False)
    if scale_x:
        normx = np.sqrt(np.dot(one, x**2) / (n - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        x = scale(x, with_std=False) / normx
    else:
        normx = np.ones(p)
    if scale_y:
        normy = np.sqrt(np.dot(one, y**2) / (n - 1))
        if np.any(normy < np.finfo(float).eps):
            raise ValueError("Some of the columns of the response matrix have zero variance.")
        y = scale(y, with_std=False) / normy
    else:
        normy = np.ones(q)
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    U, s, Vt = svds(np.dot(x.T, y), k=1)
    u = U
    v = Vt.T
    iter = 1
    dis_u = 10
    dis_v = 10
    if trace:
        print("The variables that join the set of selected variables at each step:")
    while dis_u > eps and dis_v > eps and iter <= maxstep:
        u_old = u.copy()
        v_old = v.copy()
        s = np.array(np.dot(np.dot(v.T, y.T), x) / n).reshape(-1)
        ro_d = np.array([ro_d1st(s[j], mu1, 6) for j in range(p)])
        fun_c = lambda j: np.sign(s[j]) * (abs(s[j]) > ro_d[j]) * (abs(s[j]) - ro_d[j])
        u = np.array([fun_c(j) for j in range(p)])
        u_norm = np.sqrt(np.sum(u**2))
        u_norm = 1e-04 if u_norm == 0 else u_norm
        u = u / u_norm
        dis_u = np.sqrt(np.sum((u - u_old)**2)) / np.sqrt(np.sum(u_old**2))
        what = u
        what_cut = np.where(abs(what) > 1e-04, what, 0)
        what_cut_norm = np.sqrt(np.sum(what_cut**2))
        what_cut_norm = 1e-04 if what_cut_norm == 0 else what_cut_norm
        what_dir = what_cut / what_cut_norm
        what_cut = np.where(abs(what_dir) > 1e-04, what_dir, 0)
        s = np.dot(np.dot(u.T, x.T), y) / n
        ro_d = np.array([ro_d1st(s[j], mu2, 6) for j in range(q)])
        fun_c = lambda j: np.sign(s[j]) * (abs(s[j]) > ro_d[j]) * (abs(s[j]) - ro_d[j])
        v = np.array([fun_c(j) for j in range(q)])
        v_norm = np.sqrt(np.sum(v**2))
        v_norm = 1e-04 if v_norm == 0 else v_norm
        v = v / v_norm
        dis_v = np.sqrt(np.sum((v - v_old)**2)) / np.sqrt(np.sum(v_old**2))
        what_v = v
        what_cut_v = np.where(abs(what_v) > 1e-04, what_v, 0)
        what_cut_norm = np.sqrt(np.sum(what_cut_v**2))
        what_cut_norm = 1e-04 if what_cut_norm == 0 else what_cut_norm
        what_dir = what_cut_v / what_cut_norm
        what_cut_v = np.where(abs(what_dir) > 1e-04, what_dir, 0)
        dis_u = np.sqrt(np.sum((u - u_old)**2)) / np.sqrt(np.sum(u_old**2))
        dis_v = np.sqrt(np.sum((v - v_old)**2)) / np.sqrt(np.sum(v_old**2))
        if np.sum(abs(v) <= 1e-04) == q and np.sum(abs(u) <= 1e-04) == p:
            print("Stop! The values of mu1 and mu2 are too large")
            break
        if np.sum(abs(v) <= 1e-04) != q and np.sum(abs(u) <= 1e-04) == p:
            print("Stop! The value of mu1 is too large")
            break
        if np.sum(abs(v) <= 1e-04) == q and np.sum(abs(u) <= 1e-04) != p:
            print("Stop! The value of mu2 is too large")
            break
        if trace:
            new2A_u = ip[what_cut != 0]
            new2A_v = iq[what_cut_v != 0]
            print("\n--------------------\n----- Step {} -----\n--------------------".format(iter))
            new2A_l = new2A_u
            if len(new2A_l) <= 10:
                print(', '.join(['X{}'.format(i) for i in new2A_l]))
            else:
                nlines = int(np.ceil(len(new2A_l) / 10))
                for i in range(nlines - 1):
                    print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
            new2A_l = new2A_v
            if len(new2A_l) <= 10:
                print(', '.join(['Y{}'.format(i) for i in new2A_l]))
            else:
                nlines = int(np.ceil(len(new2A_l) / 10))
                for i in range(nlines - 1):
                    print(', '.join(['Y{}'.format(i) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                print(', '.join(['Y{}'.format(i) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
        iter += 1
    what = what_cut
    what_v = what_cut_v
    new2A = ip[what != 0]
    new2A_v = iq[what_v != 0]  
    class scca:
        def __init__(self, data):
            self.x = data['x']
            self.y = data['y']
            self.loading_x = data['loading.x']
            self.loading_y = data['loading.y']
            self.variable_x = data['variable.x']
            self.variable_y = data['variable.y']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.meany = data['meany']
            self.normy = data['normy']
            self.mu1 = data['mu1']
            self.mu2 = data['mu2']
    object_dict = {
        'x': x, 
        'y': y, 
        'loading.x': what, 
        'loading.y': what_v, 
        'variable.x': new2A, 
        'variable.y': new2A_v, 
        'meanx': meanx, 
        'normx': normx, 
        'meany': meany, 
        'normy': normy, 
        'mu1': mu1, 
        'mu2': mu2, 
    }
    object = scca(object_dict)
    return object



import numpy as np
from scipy.integrate import quad
from sklearn.preprocessing import scale
from sklearn.model_selection import KFold
from numpy.linalg import svd
from numpy.linalg import solve
from numpy.linalg import norm
from scipy.optimize import fsolve
from numpy.linalg import lstsq
import itertools
from .Ispls import ispls

def ispls_cv(x, y, L, K, mu1, mu2, eps=1e-04, kappa=0.05, pen1="homogeneity", 
             pen2="magnitude", scale_x=True, scale_y=True, maxstep=50, submaxstep=10):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    if not isinstance(y, list):
        raise TypeError("y should be of list type.")
    x = [np.array(i) for i in x]
    y = [np.array(i) for i in y]
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x]
    ql = [i.shape[1] for i in y]
    p = list(set(pl))
    q = list(set(ql))   
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(q) > 1:
        raise ValueError("The dimension of data y should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    if len(q) == 1:
        q = q[0] 
    ip = np.array(range(1, p+1))
    iq = np.array(range(1, q+1))
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    meany = [np.dot(np.ones((1, nl[l])), y[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
        y[l] = scale(y[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    def y_scale(l):
        one = np.ones((1, nl[l]))
        normy = np.sqrt(np.dot(one, y[l]**2) / (nl[l] - 1))
        if np.any(normy < np.finfo(float).eps):
            raise ValueError("Some of the columns of the response matrix have zero variance.")
        return normy
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_y:
        normy = [y_scale(i) for i in range(L)]
    else:
        normy = [[1]*q[0]]*L
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    if scale_y:
        y = [(y[i] - np.mean(y[i], axis=0)) / normy[i] for i in range(L)]
    folds = [list(KFold(n_splits=K, shuffle = True).split(range(nl[l]))) for l in range(L)]
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    def c_value_homo(z, a, c, p, q, L, mu1, mu2, pen2, nl):
        def fun_s(j, l):
            z_l = z[:, (l * q):(l + 1) * q]
            a_l = a[:, l].reshape(-1, 1)
            c_j = c[j, :]
            s1 = np.dot(a_l.T, np.dot(z_l, z_l[j, :].reshape(1, -1).T)) / (nl[l]**2)
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(c_j[np.arange(len(c_j))!=l])
            elif pen2 == "sign":
                s2 = mu2 * np.sum(np.array([x / np.sqrt(x**2 + 0.5) for x in c_j[np.arange(len(c_j))!=l]])) / np.sqrt(c_j[l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        norm_c_j = np.sqrt(np.sum(c**2, axis=1))
        ro_d = ro_d1st(norm_c_j, mu1, 6)
        def fun_c(j, l):
            s_norm = np.sqrt(np.sum(result_s[j, :]**2))
            if pen2 == "magnitude":
                temp = (s_norm > ro_d[j]) * result_s[j, l] * (s_norm - ro_d[j]) / ((1/q/(nl[l]**2) + mu2 * (L - 1)) * s_norm)
            elif pen2 == "sign":
                temp = (s_norm > ro_d[j]) * result_s[j, l] * (s_norm - ro_d[j]) / ((1/q/(nl[l]**2) + mu2 * (L - 1)/(c[j, l]**2 + 0.5)) * s_norm)
            return temp
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def c_value_hetero(z, a, c, p, q, L, mu1, mu2, pen2, nl):
        def fun_mu(j, l):
            c_j = c[j, :]
            ro_j = ro(np.abs(c_j), mu1, 6)
            s_ro = np.sum(ro_j)
            mu_jl = ro_d1st(s_ro, 1, 0.5 * L * 6 * mu1**2) * ro_d1st(np.abs(c_j[l]), mu1, 6)
            return mu_jl
        result_mu = np.array([fun_mu(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        def fun_s(j, l):
            z_l = z[:, (l * q):(l + 1) * q]
            a_l = a[:, l].reshape(-1, 1)
            c_j = c[j, :]
            s1 = np.dot(a_l.T, np.dot(z_l, z_l[j, :].reshape(1, -1).T)) / (nl[l]**2)
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(c_j[np.arange(len(c_j))!=l])
            elif pen2 == "sign":
                s2 = mu2 * np.sum(np.array([x / np.sqrt(x**2 + 0.5) for x in c_j[np.arange(len(c_j))!=l]])) / np.sqrt(c_j[l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        def fun_c(j, l):
            if pen2 == "magnitude":
                temp = np.sign(result_s[j, l]) * (np.abs(result_s[j, l]) > result_mu[j, l]) * (np.abs(result_s[j, l]) - result_mu[j, l]) / (1/q/(nl[l]**2) + mu2 * (L - 1))
            elif pen2 == "sign":
                temp = np.sign(result_s[j, l]) * (np.abs(result_s[j, l]) > result_mu[j, l]) * (np.abs(result_s[j, l]) - result_mu[j, l]) / (1/q/(nl[l]**2) + mu2 * (L - 1)/(c[j, l]**2 + 0.5))
            return temp
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def c_value_homo_q1(z, a, c, p, q, L, mu1, mu2, pen2, nl):
        def fun_s(j, l):
            z_l = z[:, (l * q):(l + 1) * q]
            a_l = a[:, l].reshape(-1, 1)
            c_j = c[j, :]
            s1 = np.dot(a_l.T, np.dot(z_l, z_l[j].reshape(1, -1).T)) / (nl[l]**2)
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(c_j[np.arange(len(c_j))!=l])
            elif pen2 == "sign":
                s2 = mu2 * np.sum(np.array([x / np.sqrt(x**2 + 0.5) for x in c_j[np.arange(len(c_j))!=l]])) / np.sqrt(c_j[l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        norm_c_j = np.sqrt(np.sum(c**2, axis=1))
        ro_d = ro_d1st(norm_c_j, mu1, 6)
        def fun_c(j, l):
            s_norm = np.sqrt(np.sum(result_s[j, :]**2))
            if pen2 == "magnitude":
                temp = (s_norm > ro_d[j]) * result_s[j, l] * (s_norm - ro_d[j]) / ((1/q/(nl[l]**2) + mu2 * (L - 1)) * s_norm)
            elif pen2 == "sign":
                temp = (s_norm > ro_d[j]) * result_s[j, l] * (s_norm - ro_d[j]) / ((1/q/(nl[l]**2) + mu2 * (L - 1)/(c[j, l]**2 + 0.5)) * s_norm)
            return temp
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def c_value_hetero_q1(z, a, c, p, q, L, mu1, mu2, pen2, nl):
        def fun_mu(j, l):
            c_j = c[j, :]
            ro_j = ro(np.abs(c_j), mu1, 6)
            s_ro = np.sum(ro_j)
            mu_jl = ro_d1st(s_ro, 1, 0.5 * L * 6 * mu1**2) * ro_d1st(np.abs(c_j[l]), mu1, 6)
            return mu_jl
        result_mu = np.array([fun_mu(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        def fun_s(j, l):
            z_l = z[:, (l * q):(l + 1) * q]
            a_l = a[:, l].reshape(-1, 1)
            c_j = c[j, :]
            s1 = np.dot(a_l.T, np.dot(z_l, z_l[j].reshape(1, -1).T)) / (nl[l]**2)
            if pen2 == "magnitude":
                s2 = mu2 * np.sum(c_j[np.arange(len(c_j))!=l])
            elif pen2 == "sign":
                s2 = mu2 * np.sum(np.array([x / np.sqrt(x**2 + 0.5) for x in c_j[np.arange(len(c_j))!=l]])) / np.sqrt(c_j[l]**2 + 0.5)
            s = s1 + s2
            return s
        result_s = np.array([fun_s(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        def fun_c(j, l):
            if pen2 == "magnitude":
                temp = np.sign(result_s[j, l]) * (np.abs(result_s[j, l]) > result_mu[j, l]) * (np.abs(result_s[j, l]) - result_mu[j, l]) / (1/q/(nl[l]**2) + mu2 * (L - 1))
            elif pen2 == "sign":
                temp = np.sign(result_s[j, l]) * (np.abs(result_s[j, l]) > result_mu[j, l]) * (np.abs(result_s[j, l]) - result_mu[j, l]) / (1/q/(nl[l]**2) + mu2 * (L - 1)/(c[j, l]**2 + 0.5))
            return temp
        c = np.array([fun_c(j, l) for j in range(p) for l in range(L)]).reshape(p, L)
        return c
    def fun_k(k):
        x_train = [x[l][folds[l][k][0]] for l in range(L)]
        y_train = [y[l][folds[l][k][0]] for l in range(L)]
        x_test = [x[l][folds[l][k][1]] for l in range(L)]
        y_test = [y[l][folds[l][k][1]] for l in range(L)]
        nl_train = [len(x_t) for x_t in x_train]
        nl_test = [len(x_t) for x_t in x_test]
        return {'x_train': x_train, 'y_train': y_train, 'x_test': x_test, 
                'y_test': y_test, 'nl_train': nl_train, 'nl_test': nl_test}
    data_cv = [fun_k(k) for k in range(K)]
    def cv_fun(k):
        x_train = data_cv[k]['x_train']
        y_train = data_cv[k]['y_train']
        x_test = data_cv[k]['x_test']
        y_test = data_cv[k]['y_test']
        nl_train = data_cv[k]['nl_train']
        nl_test = data_cv[k]['nl_test']
        what = np.zeros((p, L))
        def fun_1(l):
            z_l = np.dot(np.transpose(x_train[l]), y_train[l])
            z_l = z_l / nl_train[l]
            return z_l
        z = np.hstack(np.array([fun_1(l) for l in range(L)]))
        def fun_2(l):
            z_l = z[:, (l * q):(l + 1) * q]
            m_l = np.dot(z_l, z_l.T) / q
            return m_l
        m = np.hstack(np.array([fun_2(l) for l in range(L)]))
        dis = 10
        iter = 1
        kappa2 = (1 - kappa) / (1 - 2 * kappa)
        def fun_c(l):
            return svd(np.dot(z[:, (l * q):(l + 1) * q], z[:, (l * q):(l + 1) * q].T), full_matrices=False)[0][:, 0]
        c = np.array([fun_c(l) for l in range(L)]).T
        sgn1 = np.array([np.sign(np.dot(np.dot(c[:, l].T, z[:, (l * q):(l + 1) * q]), np.dot(z[:, (l * q):(l + 1) * q].T, c[:, l]))) for l in range(L)])
        for l in range(L):
            c[:, l] = sgn1[l] * c[:, l]
        sgn2 = np.array([np.sign(np.dot(c[:, 0].T, c[:, l])) / (np.sqrt(np.sum(c[:, 0]**2)) * np.sqrt(np.sum(c[:, l]**2))) for l in range(1, L)])
        for l in range(1, L):
            c[:, l] = sgn2[l - 1] * c[:, l]
        a = c
        dis_c_iter = 10
        while dis_c_iter > eps and iter <= maxstep:
            c_iter = c.copy()
            def fun_3(l):
                def h(lambda_):
                    alpha = solve(m[:, l * p: (l + 1) * p] + lambda_ * np.eye(p), np.dot(m[:, l * p: (l + 1) * p], c[:, l]))
                    obj = np.dot(alpha.T, alpha) - 1 / kappa2**2
                    return obj
                while h(1e-04) * h(1e+12) > 0:
                    m[:, (l - 1) * p: l * p] *= 2
                    c[:, l] *= 2
                lambdas = fsolve(h, x0=1e-04)
                a_l = kappa2 * solve(m[:, l * p: (l + 1) * p] + lambdas * np.eye(p), np.dot(m[:, l * p: (l + 1) * p], c[:, l]))
                return a_l
            a = np.array([fun_3(l) for l in range(L)]).T
            subiter_c = 1
            dis_c = 10
            while dis_c > eps and subiter_c <= submaxstep:
                c_old = c.copy()
                if pen1 == "homogeneity" and q != 1:
                    c = c_value_homo(z, a, c, p, q, L, mu1, mu2, pen2, nl = nl_train)
                elif pen1 == "homogeneity" and q == 1:
                    c = c_value_homo_q1(z, a, c, p, q, L, mu1, mu2, pen2, nl = nl_train)
                elif pen1 == "heterogeneity" and q != 1:
                    c = c_value_hetero(z, a, c, p, q, L, mu1, mu2, pen2, nl = nl_train)
                elif pen1 == "heterogeneity" and q == 1:
                    c = c_value_hetero_q1(z, a, c, p, q, L, mu1, mu2, pen2, nl = nl_train)
                c_norm = np.sqrt(np.sum(c**2, axis=0))
                c_norm[c_norm == 0] = 1e-04
                c = (c.T / np.expand_dims(c_norm,axis=-1)).T
                dis = np.max(np.sqrt(np.sum((c - c_old)**2, axis=0)) / np.sqrt(np.sum(c_old**2, axis=0)))
                what = c.copy()
                what_cut = np.where(np.abs(what) > 1e-04, what, 0)
                what_cut_norm = np.sqrt(np.sum(what_cut**2, axis=0))
                what_cut_norm[what_cut_norm == 0] = 1e-04
                what_dir = (what_cut.T / np.expand_dims(what_cut_norm, axis=-1)).T
                what_cut = np.where(np.abs(what_dir) > 1e-04, what_dir, 0)
                if np.sum(np.apply_along_axis(lambda x: np.sum(np.abs(x) <= 1e-04) == p, axis=0, arr=c)) > 0:
                    print("Stop! The value of mu1 is too large")
                    break
                subiter_c += 1
            dis_c_iter = np.max(np.sqrt(np.sum((c - c_iter)**2, axis=0)) / np.sqrt(np.sum(c_iter**2, axis=0)))
            if np.sum(np.apply_along_axis(lambda x: np.sum(np.abs(x) <= 1e-04) == p, axis=0, arr=c)) > 0:
                break
            iter += 1
        def fun_fit_cv(l):
            x_l = x_train[l]
            w_l = what[:, l].reshape(-1, 1)
            t_l = np.dot(x_l, w_l)
            if np.sum(w_l == 0) != p:
                y_l = y_train[l]
                fit_l = lstsq(t_l, y_l, rcond=None)[0]
                betahat_l = np.dot(w_l, fit_l).reshape(p, q)
            else:
                betahat_l = np.zeros((p, q))
            return betahat_l
        betahat_cv = [fun_fit_cv(i) for i in range(L)]
        def get_rho(l):
            return norm(y_test[l] - np.matmul(x_test[l], betahat_cv[l]))**2
        cv_rho = sum([get_rho(l) for l in range(L)])
        return cv_rho
    mu_list = [mu1, mu2]
    for i in range(len(mu_list)):
        if not isinstance(mu_list[i], list):
            mu_list[i] = [mu_list[i]]
    mu1, mu2 = mu_list
    mu = np.array(list(itertools.product(mu1, mu2)))
    for i in range(len(mu_list)):
        if isinstance(mu_list[i], list) and len(mu_list[i]) > 1:
            mu_list[i] = np.array(mu_list[i])
        else:
            mu_list[i] = float(mu_list[i][0])
    mu1, mu2 = mu_list
    rho = []
    for loop in range(mu.shape[0]):
        mu1 = mu[loop, 0]
        mu2 = mu[loop, 1]
        cv_fun_list = []
        for k in range(K):
            cv_fun_list.append(cv_fun(k))
        rho.append(np.mean(cv_fun_list))
    index = np.argmax(rho)
    mu1_final = mu[index, 0]
    mu2_final = mu[index, 1]
    result = ispls(x, y, L, mu1 = mu1_final, mu2 = mu2_final, eps=eps, kappa = kappa, 
                   pen1 = pen1, pen2 = pen2, scale_x = scale_x, scale_y = scale_y, 
                   maxstep = maxstep, trace = False, draw = False)
    dictname = [f'Dataset {l+1}' for l in range(L)]
    folds_dict = {}
    meanx_dict = {}
    meany_dict = {}
    normx_dict = {}
    normy_dict = {}
    x_dict = {}
    y_dict = {}    
    for l in range(L):
        folds_dict.update({dictname[l]:folds[l]})
        meanx_dict.update({dictname[l]:meanx[l]})
        meany_dict.update({dictname[l]:meany[l]})
        normx_dict.update({dictname[l]:normx[l]})
        normy_dict.update({dictname[l]:normy[l]})        
        x_dict.update({dictname[l]:x[l]})
        y_dict.update({dictname[l]:y[l]})            
    class ispls_cv:
        def __init__(self, data):
            self.x = data['x']
            self.y = data['y']
            self.mu1 = data['mu1']
            self.mu2 = data['mu2']
            self.fold = data['fold']
            self.betahat = data['betahat']
            self.loading = data['loading']
            self.variable = data['variable']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.meany = data['meany']
            self.normy = data['normy']
            self.pen1 = data['pen1']
            self.pen2 = data['pen2']
            self.kappa = data['kappa']
            self.loading_trace = data['loading_trace']
    object_dict = {
        'x': x_dict, 
        'y': y_dict,
        'mu1': mu1_final, 
        'mu2': mu2_final, 
        'fold': folds_dict,
        'betahat': result.betahat,
        'loading': result.loading,
        'variable': result.variable, 
        'meanx': meanx_dict, 
        'normx': normx_dict, 
        'meany': meany_dict, 
        'normy': normy_dict, 
        'pen1': pen1, 
        'pen2': pen2, 
        'kappa': kappa, 
        'loading_trace': result.loading_trace
    }
    object = ispls_cv(object_dict)
    return object


import matplotlib.pyplot as plt
import numpy as np

def ispca_plot(x, type):
    loadings = x.eigenvector
    L = loadings.shape[1]
    p = loadings.shape[0]
    loading_trace = x.loading_trace
    iter = loading_trace.shape[1]
    if type == "path":
        for l in range(L):
            plt.figure()
            plt.plot(np.transpose(loading_trace[(l * p):((l + 1) * p), :]))
            plt.title(f"Dataset {l+1}\nConvergence path of elements in vector u")
            plt.xlabel("Number of iterations")
            plt.ylabel("Weight")
            plt.show()
    if type == "loading":
        for l in range(L):
            plt.figure()
            plt.scatter(range(1, p+1), loading_trace[(l * p):((l + 1) * p), iter-1])
            plt.title(f"The first canonical vector u of dataset {l+1}")
            plt.xlabel("Dimension")
            plt.ylabel("Value")
            plt.show()


import numpy as np
from sklearn.preprocessing import scale
from numpy.linalg import svd
import matplotlib.pyplot as plt

def preview_pls(x, y, L, scale_x=True, scale_y=True):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    if not isinstance(y, list):
        raise TypeError("y should be of list type.")
    x = [np.array(i) for i in x]
    y = [np.array(i) for i in y]  
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x]    
    ql = [i.shape[1] for i in y]  
    p = list(set(pl))
    q = list(set(ql))   
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(q) > 1:
        raise ValueError("The dimension of data y should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    if len(q) == 1:
        q = q[0]
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    meany = [np.dot(np.ones((1, nl[l])), y[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
        y[l] = scale(y[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    def y_scale(l):
        one = np.ones((1, nl[l]))
        normy = np.sqrt(np.dot(one, y[l]**2) / (nl[l] - 1))
        if np.any(normy < np.finfo(float).eps):
            raise ValueError("Some of the columns of the response matrix have zero variance.")
        return normy
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_y:
        normy = [y_scale(i) for i in range(L)]
    else:
        normy = [[1]*q[0]]*L
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    if scale_y:
        y = [(y[i] - np.mean(y[i], axis=0)) / normy[i] for i in range(L)]
    def fun_1(l):
        z_l = np.dot(np.transpose(x[l]), y[l])
        z_l = z_l / nl[l]
        return z_l
    z = np.hstack(np.array([fun_1(l) for l in range(L)]))
    def fun_c(l):
        return svd(np.dot(z[:, (l * q):(l + 1) * q], z[:, (l * q):(l + 1) * q].T), full_matrices=False)[0][:, 0]
    c = np.array([fun_c(l) for l in range(L)]).T
    what = c
    dictname = [f'Dataset {l+1}' for l in range(L)]
    meanx_dict = {}
    meany_dict = {}
    normx_dict = {}
    normy_dict = {}
    x_dict = {}
    y_dict = {}    
    for l in range(L):
        meanx_dict.update({dictname[l]:meanx[l]})
        meany_dict.update({dictname[l]:meany[l]})
        normx_dict.update({dictname[l]:normx[l]})
        normy_dict.update({dictname[l]:normy[l]})        
        x_dict.update({dictname[l]:x[l]})
        y_dict.update({dictname[l]:y[l]})
    for l in range(L):
        plt.figure()
        plt.scatter(range(1, p+1), what[:, l])
        plt.title('Dataset {}\nThe first direction vector'.format(l+1))
        plt.xlabel('Dimension')
        plt.ylabel('Value')
        plt.show()
    class preview_pls:
        def __init__(self, data):
            self.x = data['x']
            self.y = data['y']
            self.loading_x = data['loading']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.meany = data['meany']
            self.normy = data['normy']
    object_dict = {
        'x': x_dict, 
        'y': y_dict, 
        'loading': what, 
        'meanx': meanx_dict, 
        'normx': normx_dict, 
        'meany': meany_dict, 
        'normy': normy_dict, 
    }
    object = preview_pls(object_dict)
    return object


import numpy as np
from scipy.integrate import quad
from sklearn.preprocessing import scale
from scipy.sparse.linalg import svds

def meta_spca(x, L, mu1, eps=1e-04, scale_x=True, maxstep=50, trace=False):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    x = [np.array(i) for i in x] 
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x] 
    p = list(set(pl))
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    ip = np.array(range(1, p+1))
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    z = np.zeros((p, p))
    for l in range(L):
        z += (nl[l] - 1) * np.cov(x[l], rowvar = False)
    z /= (np.sum(nl) - L)
    u, s, vt = svds(z, k=1)
    v = u
    lambda_ = np.zeros(p)
    dis_u = 10
    iter = 1
    if trace:
        print("The variables that join the set of selected variables at each step:")
    while dis_u > eps and iter <= maxstep:
        u_old = np.copy(u)
        v_old = np.copy(v)
        s = np.array(np.dot(v.T, z)).reshape(-1)
        ro_d = [ro_d1st(s[j], mu1, 6) for j in range(p)]
        fun_c = lambda j: np.sign(s[j]) * (abs(s[j]) > ro_d[j]) * (abs(s[j]) - ro_d[j])
        u = np.array([fun_c(j) for j in range(p)])
        if np.sum(abs(u) <= 1e-04) == p:
            print("The value of mu1 is too large")
            break
        v = np.dot(z, u) / np.sqrt(np.sum(np.square(np.dot(z, u))))
        u_norm = np.sqrt(np.sum(np.square(u)))
        u_norm = 1e-04 if u_norm == 0 else u_norm
        u = u / u_norm
        dis_u = np.sqrt(np.sum(np.square(u - u_old))) / np.sqrt(np.sum(np.square(u_old)))
        what = u
        what_cut = np.where(abs(what) > 1e-04, what, 0)
        what_cut_norm = np.sqrt(np.sum(np.square(what_cut)))
        what_cut_norm = 1e-04 if what_cut_norm == 0 else what_cut_norm
        what_dir = what_cut / what_cut_norm
        what_cut = np.where(abs(what_dir) > 1e-04, what_dir, 0)
        if np.sum(abs(u) <= 1e-04) == p:
            print("The value of mu1 is too large")
            break
        if trace:
            new2A = ip[what != 0]
            print("\n--------------------\n----- Step {} -----\n--------------------".format(iter))
            new2A_l = new2A
            if len(new2A_l) <= 10:
                print(', '.join(['X{}'.format(i) for i in new2A_l]))
            else:
                nlines = int(np.ceil(len(new2A_l) / 10))
                for i in range(nlines - 1):
                    print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
        iter += 1
    what = what_cut
    new2A = ip[what != 0]
    eigenvalue = [what.T.dot(np.cov(x[l], rowvar = False)).dot(what) for l in range(L)]
    comp = [x[l].dot(what) for l in range(L)]
    dictname = [f'Dataset {l+1}' for l in range(L)]
    new2A_dict = {}
    meanx_dict = {}
    normx_dict = {}
    x_dict = {}
    eigenvalue_dict = {}
    comp_dict = {}
    for l in range(L):
        new2A_dict.update({dictname[l]:new2A[l]})
        meanx_dict.update({dictname[l]:meanx[l]})
        normx_dict.update({dictname[l]:normx[l]})      
        x_dict.update({dictname[l]:x[l]})
        eigenvalue_dict.update({dictname[l]:eigenvalue[l]})
        comp_dict.update({dictname[l]:comp[l]})
    class meta_spca:
        def __init__(self, data):
            self.x = data['x']
            self.eigenvalue = data['eigenvalue']
            self.eigenvector = data['eigenvector']
            self.component = data['component']
            self.variable = data['variable']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.mu1 = data['mu1']
    object_dict = {
        'x': x_dict, 
        'eigenvalue': eigenvalue_dict, 
        'eigenvector': what, 
        'component': comp_dict, 
        'variable': new2A, 
        'meanx': meanx_dict,
        'normx': normx_dict, 
        'mu1': mu1, 
    }
    object = meta_spca(object_dict)
    return object


import numpy as np
import pandas as pd
from scipy.integrate import quad
from sklearn.preprocessing import scale
from numpy.linalg import svd
from numpy.linalg import solve
from scipy.optimize import root_scalar
from numpy.linalg import lstsq

def meta_spls(x, y, L, mu1, eps=1e-04, kappa=0.05, scale_x=True, scale_y=True, 
              maxstep=50, trace=False):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    if not isinstance(y, list):
        raise TypeError("y should be of list type.")
    x = [np.array(i) for i in x]
    y = [np.array(i) for i in y]
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x]
    ql = [i.shape[1] for i in y]
    p = list(set(pl))
    q = list(set(ql))   
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(q) > 1:
        raise ValueError("The dimension of data y should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    if len(q) == 1:
        q = q[0] 
    ip = np.array(range(1, p+1))
    iq = np.array(range(1, q+1))
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    meany = [np.dot(np.ones((1, nl[l])), y[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
        y[l] = scale(y[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    def y_scale(l):
        one = np.ones((1, nl[l]))
        normy = np.sqrt(np.dot(one, y[l]**2) / (nl[l] - 1))
        if np.any(normy < np.finfo(float).eps):
            raise ValueError("Some of the columns of the response matrix have zero variance.")
        return normy
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_y:
        normy = [y_scale(i) for i in range(L)]
    else:
        normy = [[1]*q[0]]*L
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    if scale_y:
        y = [(y[i] - np.mean(y[i], axis=0)) / normy[i] for i in range(L)]
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    def fun_1(l):
        z_l = np.dot(np.transpose(x[l]), y[l])
        return z_l
    zz = [fun_1(l) for l in range(L)]
    z = np.zeros((p, q))
    for l in range(L):
        z += zz[l]
    z /= np.sum(nl)
    c = svd(np.dot(z, z.T), full_matrices=False)[0][:, :1]
    a = c
    iter = 1
    dis = 10
    kappa2 = (1 - kappa)/(1 - 2 * kappa)
    if trace:
        print("The variables that join the set of selected variables at each step:")
    while dis > eps and iter <= maxstep:
        c_old = c.copy()
        a_old = a.copy()
        m = np.dot(z, z.T) / q
        def h(lambda_):
            alpha = solve(m + lambda_ * np.eye(p), np.dot(m, c))
            obj = np.dot(alpha.T, alpha) - 1/kappa2**2
            return obj
        while h(1e-04) * h(1e+12) > 0:
            m = 2 * m
            c = 2 * c
        lambdas = root_scalar(h, method='brentq', bracket=[1e-04, 1e+12]).root
        a = kappa2 * solve(m + lambdas * np.eye(p), np.dot(m, c))
        s = np.array(np.dot(a.T, np.dot(z, z.T)) / np.sum(nl)).reshape(-1)
        ro_d = ro_d1st(s, mu1, 6)
        fun_c = lambda j: np.sign(s[j]) * (abs(s[j]) > ro_d[j]) * (abs(s[j]) - ro_d[j])
        c = np.array([fun_c(j) for j in range(p)])
        c_norm = np.sqrt(np.sum(np.array(c)**2))
        c_norm = 1e-04 if c_norm == 0 else c_norm
        c = c/c_norm
        dis = np.sqrt(np.sum((c - c_old)**2))/np.sqrt(np.sum(c_old**2))
        what = c.copy()
        what_cut = np.where(np.abs(what) > 1e-04, what, 0)
        what_cut_norm = np.sqrt(np.sum(what_cut**2, axis=0))
        what_cut_norm = 1e-04 if what_cut_norm == 0 else what_cut_norm
        what_dir = (what_cut.T / np.expand_dims(what_cut_norm, axis=-1)).T
        what_cut = np.where(np.abs(what_dir) > 1e-04, what_dir, 0)
        if np.sum(np.abs(c) <= 1e-04) == p:
            print("The value of mu1 is too large")
            break
        if trace:
            new2A = ip[what_cut != 0]
            print("\n--------------------\n----- Step {} -----\n--------------------".format(iter))
            new2A_l = new2A
            if len(new2A_l) <= 10:
                print("DataSet {}:".format(l+1))
                print(', '.join(['X{}'.format(i) for i in new2A_l]))
            else:
                print("DataSet {}:".format(l+1))
                nlines = int(np.ceil(len(new2A_l) / 10))
                for i in range(nlines - 1):
                    print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * i):(10 * (i + 1))]]))
                print(', '.join(['X{}'.format(i) for i in new2A_l[(10 * (nlines - 1)):len(new2A_l)]]))
        iter += 1
    what = what_cut
    new2A = ip[what != 0]
    betahat = np.zeros((p, q * L))
    def fun_fit(l):
        x_l = x[l]
        w_l = np.matrix(what).reshape(-1, 1)
        t_l = np.dot(x_l, w_l)
        if np.sum(w_l == 0) != p:
            y_l = y[l]
            fit_l = lstsq(t_l, y_l, rcond=None)[0]
            betahat_l = np.dot(w_l, fit_l).reshape(p, q)
        else:
            betahat_l = np.zeros((p, q))
        if x[l].shape[1] > 1:
            betahat_l = pd.DataFrame(betahat_l, index=np.arange(1, p+1))
        if q > 1 and y[l].shape[1] > 1:
            betahat_l.columns = np.arange(1, q+1)
        return betahat_l
    betahat = [fun_fit(l) for l in range(L)]
    dictname = [f'Dataset {l+1}' for l in range(L)]
    betahat_dict = {}
    meanx_dict = {}
    meany_dict = {}
    normx_dict = {}
    normy_dict = {}
    x_dict = {}
    y_dict = {}    
    for l in range(L):
        betahat_dict.update({dictname[l]:betahat[l]})
        meanx_dict.update({dictname[l]:meanx[l]})
        meany_dict.update({dictname[l]:meany[l]})
        normx_dict.update({dictname[l]:normx[l]})
        normy_dict.update({dictname[l]:normy[l]})        
        x_dict.update({dictname[l]:x[l]})
        y_dict.update({dictname[l]:y[l]}) 
    class meta_spls:
        def __init__(self, data):
            self.x = data['x']
            self.y = data['y']
            self.betahat = data['betahat']
            self.loading = data['loading']
            self.variable = data['variable']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.meany = data['meany']
            self.normy = data['normy']
            self.mu1 = data['mu1']
    object_dict = {
        'x': x_dict, 
        'y': y_dict, 
        'betahat': betahat_dict, 
        'loading': what, 
        'variable': new2A,
        'meanx': meanx_dict, 
        'normx': normx_dict, 
        'meany': meany_dict, 
        'normy': normy_dict, 
        'mu1': mu1
    }
    object = meta_spls(object_dict)
    return object


import matplotlib.pyplot as plt
import numpy as np

def iscca_plot(x, type):
    loading_x = x.loading_x
    loading_y = x.loading_y
    L = loading_x.shape[1]
    p = loading_x.shape[0]
    q = loading_y.shape[0]
    loading_trace_u = x.loading_trace_u
    loading_trace_v = x.loading_trace_v
    iter_u = loading_trace_u.shape[1]
    iter_v = loading_trace_v.shape[1]
    if type == "path":
        for l in range(L):
            plt.figure(figsize=(10, 5))
            plt.subplot(1, 2, 1)
            plt.plot(np.transpose(loading_trace_u[(l * p):((l + 1) * p), :]))
            plt.title(f"Dataset {l+1}\nConvergence path of elements in vector u")
            plt.xlabel("Number of iterations")
            plt.ylabel("Weight")
            plt.subplot(1, 2, 2)
            plt.plot(np.transpose(loading_trace_v[(l * q):((l + 1) * q), :]))
            plt.title(f"Dataset {l+1}\nConvergence path of elements in vector v")
            plt.xlabel("Number of iterations")
            plt.ylabel("Weight")
            plt.tight_layout()
            plt.show()
    if type == "loading":
        for l in range(L):
            plt.figure(figsize=(10, 5))
            plt.subplot(1, 2, 1)
            plt.scatter(range(1, p+1), loading_trace_u[(l * p):((l + 1) * p), iter_u-1])
            plt.title(f"The first canonical vector u of dataset {l+1}")
            plt.xlabel("Dimension")
            plt.ylabel("Value")
            plt.subplot(1, 2, 2)
            plt.scatter(range(1, p+1), loading_trace_v[(l * p):((l + 1) * p), iter_v-1])
            plt.title(f"The first canonical vector v of dataset {l+1}")
            plt.xlabel("Dimension")
            plt.ylabel("Value")
            plt.tight_layout()
            plt.show()


import numpy as np
import pandas as pd
from scipy.integrate import quad
from sklearn.preprocessing import scale
from numpy.linalg import solve
from numpy.linalg import svd
from scipy.optimize import root_scalar
from numpy.linalg import lstsq

def spls(x, y, mu1, eps=1e-04, kappa=0.05, scale_x=True, scale_y=True, maxstep=50, trace=False):
    x = np.asmatrix(x)
    y = np.asmatrix(y)
    nx, p = x.shape
    ny, q = y.shape
    if nx != ny:
        raise ValueError("The rows of data x and data y should be consistent.") 
    n = nx
    ip = np.arange(1,p+1)
    one = np.ones((1, n))
    meany = np.dot(one, y) / n
    y = scale(y, with_std=False)
    meanx = np.dot(one, x) / n
    x = scale(x, with_std=False)
    if scale_x:
        normx = np.sqrt(np.dot(one, x**2) / (n - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        x = scale(x, with_std=False) / normx
    else:
        normx = np.ones(p)
    if scale_y:
        normy = np.sqrt(np.dot(one, y**2) / (n - 1))
        if np.any(normy < np.finfo(float).eps):
            raise ValueError("Some of the columns of the response matrix have zero variance.")
        y = scale(y, with_std=False) / normy
    else:
        normy = np.ones(q)
    what = np.zeros((p, 1))
    if trace:
        print("The variables that join the set of selected variables at each step:")
    z = np.dot(x.T, y)
    z = z / n
    def ro(x, mu, alpha):
        def f(x):
            return mu * np.where(x/(mu * alpha) < 1, 1 - x/(mu * alpha), 0)
        if np.isscalar(x):
            r = quad(f, 0, x)[0]
        else:
            r = np.array([quad(f, 0, xi)[0] for xi in x])
        return r
    def ro_d1st(x, mu, alpha):
        r = mu * (1 > x/(mu * alpha)) * (1 - x/(mu * alpha))
        return r
    c = svd(np.dot(z,z.T), full_matrices=False)[0][:, 0]
    a = c
    iter = 1
    dis = 10
    kappa2 = (1 - kappa) / (1 - 2 * kappa)
    loading_trace = np.zeros((p, maxstep))
    while dis > eps and iter <= maxstep:
        c_old = c
        a_old = a
        m = np.dot(z, z.T) / q
        def h(lambda_):
            alpha = solve(m + lambda_ * np.eye(p), np.dot(m, c))
            obj = np.dot(alpha.T, alpha) - 1/kappa2**2
            return obj
        while h(1e-04) * h(1e+12) > 0:
            m = 2 * m
            c = 2 * c
        lambdas = root_scalar(h, method='brentq', bracket=[1e-04, 1e+12]).root
        a = kappa2 * solve(m + lambdas * np.eye(p), np.dot(m, c))
        s = np.array(np.dot(a.T, np.dot(z, z.T)) / np.sum(n)).reshape(-1)
        ro_d = ro_d1st(s, mu1, 6)
        fun_c = lambda j: np.sign(s[j]) * (abs(s[j]) > ro_d[j]) * (abs(s[j]) - ro_d[j])
        c = np.array([fun_c(j) for j in range(p)])
        c_norm = np.sqrt(sum(np.array(c)**2))
        c_norm = 1e-04 if c_norm == 0 else c_norm
        c = c / c_norm
        dis = np.sqrt(sum((c - c_old)**2)) / np.sqrt(sum(np.array(c_old)**2))
        what = c
        what_cut = np.where(abs(what) > 1e-04, what, 0)
        what_cut_norm = np.sqrt(sum(what_cut**2))
        what_cut_norm = 1e-04 if what_cut_norm == 0 else what_cut_norm
        what_dir = (what_cut.T / np.expand_dims(what_cut_norm, axis=-1)).T
        what_cut = np.where(abs(what_dir) > 1e-04, what_dir, 0)
        loading_trace[:, iter-1] = what_cut
        if sum(abs(c) <= 1e-04) == p:
            print("The value of mu1 is too large")
            break
        if trace:
            new2A = ip[what_cut != 0]
            print("\n--------------------\n----- Step {} -----\n--------------------".format(iter))
            if len(new2A) <= 10:
                print(', '.join(['X{}'.format(i) for i in new2A]))
            else:
                nlines = int(np.ceil(len(new2A) / 10))
                for i in range(nlines - 1):
                    print(', '.join(['X{}'.format(i) for i in new2A[(10 * i):(10 * (i + 1))]]))
                print(', '.join(['X{}'.format(i) for i in new2A[(10 * (nlines - 1)):len(new2A)]]))
        iter += 1
    loading_trace = loading_trace[:, :iter-1]
    what = what_cut
    new2A = ip[what != 0]
    betahat = np.zeros((p, q))
    w = what
    t_hat = np.dot(x, w)
    if sum(w == 0) != p:
        t_hat = t_hat.reshape(-1, 1)
        betahat = np.dot(np.expand_dims(w, axis=-1), lstsq(t_hat, y, rcond=None)[0]).reshape(p, q)
    else:
        betahat = np.zeros((p, q))
    if x.shape[1] > 1:
        betahat = pd.DataFrame(betahat, index=range(1, p + 1))
    if q > 1 and y.shape[1] > 1:
        betahat.columns = range(1, q + 1)
    class spls:
        def __init__(self, data):
            self.x = data['x']
            self.y = data['y']
            self.betahat = data['betahat']
            self.loading = data['loading']
            self.variable = data['variable']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.meany = data['meany']
            self.normy = data['normy']
            self.mu1 = data['mu1']
    object_dict = {
        'x': x, 
        'y': y, 
        'betahat': betahat, 
        'loading': what, 
        'variable': new2A,
        'meanx': meanx, 
        'normx': normx, 
        'meany': meany, 
        'normy': normy, 
        'mu1': mu1
    }
    object = spls(object_dict)
    return object



from setuptools import setup, find_packages

setup(
    name='iSFun',
    version='1.1.0',
    description='''
    The implement of integrative analysis methods based on a two-part penalization, 
    which realizes dimension reduction analysis and mining the heterogeneity and association of multiple studies with compatible designs. 
    The software package provides the integrative analysis methods including integrative sparse principal component analysis (Fang et al., 2018), 
    integrative sparse partial least squares (Liang et al., 2021) and integrative sparse canonical correlation analysis, 
    as well as corresponding individual analysis and meta-analysis versions. 
    References: (1) Fang, K., Fan, X., Zhang, Q., and Ma, S. (2018). Integrative sparse principal component analysis. Journal of Multivariate Analysis, <doi:10.1016/j.jmva.2018.02.002>. 
    (2) Liang, W., Ma, S., Zhang, Q., and Zhu, T. (2021). Integrative sparse partial least squares. Statistics in Medicine, <doi:10.1002/sim.8900>.',
    ''',
    author='Kuangnan Fang, Rui Ren, Qingzhao Zhang, Shuangge Ma',
    author_email='xmurr@stu.xmu.edu.cn',
    license='MIT',
    packages=find_packages(),
    include_package_data = True,
    package_data={'isfun_data': ['*.pkl'],},
    install_requires=[
        'numpy',
        'pandas',
        'scikit_learn',
        'scipy',
        'seaborn',
        'setuptools',
        'matplotlib'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
    ],
)



import numpy as np
import pandas as pd
from sklearn.preprocessing import scale
from scipy.sparse.linalg import svds
import matplotlib.pyplot as plt

def preview_cca(x, y, L, scale_x=True, scale_y=True):
    if not isinstance(x, list):
        raise TypeError("x should be of list type.")
    if not isinstance(y, list):
        raise TypeError("y should be of list type.")
    x = [np.array(i) for i in x]
    y = [np.array(i) for i in y]  
    nl = [i.shape[0] for i in x]
    pl = [i.shape[1] for i in x]    
    ql = [i.shape[1] for i in y]  
    p = list(set(pl))
    q = list(set(ql))   
    if len(p) > 1:
        raise ValueError("The dimension of data x should be consistent among different datasets.")
    if len(q) > 1:
        raise ValueError("The dimension of data y should be consistent among different datasets.")
    if len(p) == 1:
        p = p[0]
    if len(q) == 1:
        q = q[0]
    meanx = [np.dot(np.ones((1, nl[l])), x[l])/nl[l] for l in range(L)]
    meany = [np.dot(np.ones((1, nl[l])), y[l])/nl[l] for l in range(L)]
    for l in range(L):
        x[l] = scale(x[l], with_std=False)
        y[l] = scale(y[l], with_std=False)
    def x_scale(l):
        one = np.ones((1, nl[l]))
        normx = np.sqrt(np.dot(one, x[l]**2) / (nl[l] - 1))
        if np.any(normx < np.finfo(float).eps):
            raise ValueError("Some of the columns of the predictor matrix have zero variance.")
        return normx
    def y_scale(l):
        one = np.ones((1, nl[l]))
        normy = np.sqrt(np.dot(one, y[l]**2) / (nl[l] - 1))
        if np.any(normy < np.finfo(float).eps):
            raise ValueError("Some of the columns of the response matrix have zero variance.")
        return normy
    if scale_x:
        normx = [x_scale(i) for i in range(L)]
    else:
        normx = [[1]*p[0]]*L   
    if scale_y:
        normy = [y_scale(i) for i in range(L)]
    else:
        normy = [[1]*q[0]]*L
    if scale_x:
        x = [(x[i] - np.mean(x[i], axis=0)) / normx[i] for i in range(L)]
    if scale_y:
        y = [(y[i] - np.mean(y[i], axis=0)) / normy[i] for i in range(L)]
    def fun_1(l):
        Z_l = svds(np.dot(x[l].T, y[l]), k=1)
        u_l = Z_l[0]
        return u_l
    U = np.hstack(np.array([fun_1(l) for l in range(L)]))
    def fun_2(l):
        Z_l = svds(np.dot(x[l].T, y[l]), k=1)
        v_l = Z_l[2]
        return v_l
    V = np.vstack(np.array([fun_2(l) for l in range(L)])).T
    what_u = U
    what_v = V
    dictname = [f'Dataset {l+1}' for l in range(L)]
    meanx_dict = {}
    meany_dict = {}
    normx_dict = {}
    normy_dict = {}
    x_dict = {}
    y_dict = {}    
    for l in range(L):
        meanx_dict.update({dictname[l]:meanx[l]})
        meany_dict.update({dictname[l]:meany[l]})
        normx_dict.update({dictname[l]:normx[l]})
        normy_dict.update({dictname[l]:normy[l]})        
        x_dict.update({dictname[l]:x[l]})
        y_dict.update({dictname[l]:y[l]})
    what_u_df = pd.DataFrame(what_u, index = [str(i+1) for i in range(p)], columns = dictname)
    what_v_df = pd.DataFrame(what_v, index = [str(i+1) for i in range(p)], columns = dictname)
    def plot_loading(order, U, V, p, q):
        for l in order:
            fig, axs = plt.subplots(1, 2)
            axs[0].scatter(np.arange(1, p+1), U[:, l-1])
            axs[0].set_title(f'Dataset {l}\nThe first canonical vector u')
            axs[0].set_xlabel('Dimension')
            axs[0].set_ylabel('Value')   
            axs[1].scatter(np.arange(1, q+1), V[:, l-1])
            axs[1].set_title(f'Dataset {l}\nThe first canonical vector v')
            axs[1].set_xlabel('Dimension')
            axs[1].set_ylabel('Value') 
            plt.show()
    plot_loading(order = np.arange(1, L+1), U=U, V=V, p=p, q=q)
    class preview_cca:
        def __init__(self, data):
            self.x = data['x']
            self.y = data['y']
            self.loading_x = data['loading.x']
            self.loading_y = data['loading.y']
            self.meanx = data['meanx']
            self.normx = data['normx']
            self.meany = data['meany']
            self.normy = data['normy']
    object_dict = {
        'x': x_dict, 
        'y': y_dict, 
        'loading.x': what_u_df, 
        'loading.y': what_v_df, 
        'meanx': meanx_dict, 
        'normx': normx_dict, 
        'meany': meany_dict, 
        'normy': normy_dict, 
    }
    object = preview_cca(object_dict)
    return object
